#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse
import numpy as np
from scipy import sparse, io
import pandas as pd
import anndata as ad
from deeptools import parserCommon

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei import ReadCounter as countR
from sincei import ParserCommon

old_settings = np.seterr(all="ignore")


def parseArguments(args=None):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
            ``scCountReads`` computes the read coverages per cell barcode for genomic regions in the provided BAM file(s).
            The analysis can be performed for the entire genome by running the program in 'bins' mode.
            If you want to count the read coverage for specific regions only, use the ``features`` mode instead.
            The standard output of ``scCountReads`` is a ".h5ad" file with counts, along with rowName (features) and colNames (cell barcodes).

            A detailed sub-commands help is available by typing:

              scCountReads bins -h

              scCountReads features -h

          """,
        epilog="example usages:\n"
        "scCountReads bins --bamfiles file1.bam file2.bam --barcodes whitelist.txt -o results \n\n"
        "scCountReads features --BED selection.bed --bamfiles file1.bam file2.bam --barcodes whitelist.txt \n"
        "-o results"
        " \n\n",
        conflict_handler="resolve",
    )

    subparsers = parser.add_subparsers(
        title="commands",
        dest="command",
        description="subcommands",
        help="subcommands",
        metavar="",
    )

    read_args = ParserCommon.readOptions(suppress_args=["filterRNAstrand"])
    filter_args = ParserCommon.filterOptions()
    other_args = ParserCommon.otherOptions()

    # bins mode options
    subparsers.add_parser(
        "bins",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[
            ParserCommon.inputOutputOptions(
                opts=["bamfiles", "barcodes", "outFilePrefix", "BED"],
                requiredOpts=["barcodes", "outFilePrefix"],
                suppress_args=["BED"],
            ),
            parserCommon.gtf_options(suppress=True),
            ParserCommon.bamOptions(default_opts={"binSize": 10000, "distanceBetweenBins": 0}),
            read_args,
            filter_args,
            get_args(),
            other_args,
        ],
        help="The reads are counted in bins of equal size. The bin size and distance between bins can be adjusted.",
        add_help=False,
        usage="%(prog)s -bs 100000 --bamfiles file1.bam file2.bam -o results \n",
    )

    # BED file arguments
    subparsers.add_parser(
        "features",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[
            ParserCommon.inputOutputOptions(
                opts=["bamfiles", "barcodes", "outFilePrefix", "BED"],
                requiredOpts=["barcodes", "outFilePrefix", "BED"],
            ),
            parserCommon.gtf_options(),
            ParserCommon.bamOptions(
                suppress_args=["binSize", "distanceBetweenBins"],
                default_opts={"binSize": 10000, "distanceBetweenBins": 0},
            ),
            read_args,
            filter_args,
            get_args(),
            other_args,
        ],
        help="The user provides a BED/GTF file containing all regions "
        "that should be counted. A common use would be to count scRNA-seq reads on Genes.",
        usage="%(prog)s --BED selection.bed --bamfiles file1.bam file2.bam --barcodes whitelist.txt -o results\n",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)
    optional = parser.add_argument_group("Misc arguments")
    optional.add_argument(
        "--genomeChunkSize",
        type=int,
        default=None,
        help="Manually specify the size of the genome provided to each processor. "
        "The default value of None specifies that this is determined by read "
        "density of the BAM file.",
    )

    optional.add_argument(
        "--outFileFormat",
        type=str,
        default="h5ad",
        choices=["h5ad", "mtx"],
        help="Output file format. Default is to write an anndata object of name "
        "<prefix>.h5ad, which can either be opened in scanpy, or by downstream tools. "
        '"mtx" refers to the MatrixMarket sparse-matrix format. The output in this case would be '
        "<prefix>.counts.mtx, along with <prefix>.rownames.txt and <prefix>.colnames.txt",
    )

    return parser


def main(args=None):
    """
    1. get read counts at different positions either
    all of same length or from genomic regions from the BED file

    2. save data for further plotting
    """
    args, newlabels = ParserCommon.validateInputs(parseArguments().parse_args(args))
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    if "BED" in args:
        bed_regions = args.BED
    else:
        bed_regions = None

    ## create row/colNames
    if args.outFileFormat == "mtx":
        mtxFile = args.outFilePrefix + ".counts.mtx"
        rowNamesFile = args.outFilePrefix + ".rownames.txt"
        colNamesFile = args.outFilePrefix + ".colnames.txt"
    else:
        rowNamesFile = None

    stepSize = args.binSize + args.distanceBetweenBins
    c = countR.CountReadsPerBin(
        args.bamfiles,
        binLength=args.binSize,
        stepSize=stepSize,
        barcodes=args.barcodes,
        cellTag=args.cellTag,
        groupTag=args.groupTag,
        groupLabels=newlabels,
        motifFilter=args.motifFilter,
        genome2bit=args.genome2bit,
        GCcontentFilter=args.GCcontentFilter,
        numberOfSamples=None,
        genomeChunkSize=args.genomeChunkSize,
        numberOfProcessors=args.numberOfProcessors,
        verbose=args.verbose,
        region=args.region,
        bedFile=bed_regions,
        blackListFileName=args.blackListFileName,
        extendReads=args.extendReads,
        minMappingQuality=args.minMappingQuality,
        duplicateFilter=args.duplicateFilter,
        center_read=args.centerReads,
        samFlag_include=args.samFlagInclude,
        samFlag_exclude=args.samFlagExclude,
        minFragmentLength=args.minFragmentLength,
        maxFragmentLength=args.maxFragmentLength,
        zerosToNans=False,
        out_file_for_raw_data=rowNamesFile,
    )

    num_reads_per_bin, regionList = c.run(allArgs=args)

    sys.stderr.write("Number of bins/features " "found: {}\n".format(num_reads_per_bin.shape[0]))

    if num_reads_per_bin.shape[0] < 1:
        exit(
            "ERROR: too few non zero bins/features found.\n"
            "If using --region please check that this "
            "region is covered by reads.\n"
        )

    ## write mtx/rownames if asked
    if args.outFileFormat == "mtx":
        f = open(colNamesFile, "w")
        f.write("\n".join(newlabels))
        f.write("\n")
        f.close()
        ## write the matrix as .mtx
        sp = sparse.csr_matrix(num_reads_per_bin)
        io.mmwrite(mtxFile, sp, field="integer")
    else:
        # write anndata
        adata = ad.AnnData(num_reads_per_bin.T)
        adata.obs = pd.DataFrame(
            {
                "sample": [x.split("::")[-2] for x in newlabels],
                "barcodes": [x.split("::")[-1] for x in newlabels],
            },
            index=newlabels,
        )

        rows = list(regionList)

        adata.var = pd.DataFrame(
            {
                "chrom": [x.split("_")[0] for x in rows],
                "start": [x.split("_")[1] for x in rows],
                "end": [y.split("::")[0] for y in [x.split("_")[2] for x in rows]],
                "name": [x.split("::")[1] for x in rows],
            },
            index=rows,
        )

        # export as h5ad
        adata.write_h5ad(args.outFilePrefix + ".h5ad")



import argparse
import os
from deeptools.utilities import smartLabels
from sincei._version import __version__


def inputOutputOptions(args=None, opts=None, requiredOpts=[], suppress_args=None):
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group("Input/Output options")

    ## inputs
    if "h5adfile" in opts:
        group.add_argument(
            "--input",
            "-i",
            metavar="H5AD",
            help="Input file in .h5ad format",
            required=True,
        )
    elif "h5adfiles" in opts:
        group.add_argument(
            "--input",
            "-i",
            metavar="H5AD",
            help="List of .h5ad files separated by spaces.",
            nargs="+",
            required=True,
        )

    if "h5mufile" in opts:
        group.add_argument(
            "--input",
            "-i",
            metavar="H5MU",
            help="Input file in .h5mu format",
            required=True,
        )

    elif "bamfiles" in opts:
        group.add_argument(
            "--bamfiles",
            "-b",
            metavar="FILE1 FILE2",
            help="List of indexed bam files separated by spaces.",
            nargs="+",
            required=True,
        )
    elif "bamfile" in opts:
        group.add_argument("--bamfile", "-b", metavar="FILE", help="Indexed BAM file", required=True)

    if "whitelist" in opts:
        group.add_argument(
            "--whitelist",
            "-w",
            help="A single-column file containing the whitelist of barcodes to be used",
            metavar="TXT",
            default=None,
            required=True if "whitelist" in requiredOpts else False,
        )
    elif "barcodes" in opts:
        group.add_argument(
            "--barcodes",
            "-bc",
            help="A single-column file containing barcodes (whitelist) to be used for the analysis.",
            metavar="TXT",
            required=True if "barcodes" in requiredOpts else False,
        )

    if "groupInfo" in opts:
        group.add_argument(
            "--groupInfo",
            "-i",
            help="A 3-column tsv file with Cell grouping information in the "
            "format: `sample, barcode, group`. Coverages will be "
            "computed per group.",
            metavar="TXT file",
            required=True,
        )

    if "BED" in opts:
        group.add_argument(
            "--BED",
            help=show_or_hide(
                "Limits the coverage analysis to " "the regions specified in these files.",
                "BED",
                suppress_args,
            ),
            metavar="FILE1.bed FILE2.bed",
            nargs="+",
            required=True if "BED" in requiredOpts else False,
        )

    ## outputs
    if "outFilePrefix" in opts:
        group.add_argument(
            "--outFilePrefix",
            "-o",
            help="Output file name prefix.",
            metavar="FILEPREFIX",
            type=str,
            required=True if "outFilePrefix" in requiredOpts else False,
        )

    elif "outFile" in opts:
        group.add_argument(
            "--outFile",
            "-o",
            type=str,
            help="The file to write results to. For `scFilterStats`, `scFilterBarcodes` "
            "and `scJSD`, the output file is a .txt file. For other tools, the output file is "
            "an updated .h5ad object with the result of the requested operation. ",
            required=True if "outFile" in requiredOpts else False,
        )

    return parser


def otherOptions(args=None):
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group("Other options")

    group.add_argument("--help", "-h", action="help", help="show this help message and exit")

    group.add_argument("--verbose", "-v", help="Set to see processing messages.", action="store_true")

    group.add_argument("--version", action="version", version="%(prog)s {}".format(__version__))

    return parser


def plotOptions(args=None):
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group("Plot options")

    group.add_argument(
        "--plotWidth",
        default=10,
        type=float,
        help="Output plot width (in cm). (Default: %(default)s)",
    )

    group.add_argument(
        "--plotHeight",
        default=10,
        type=float,
        help="Output plot height (in cm). (Default: %(default)s)",
    )

    group.add_argument(
        "--plotFileFormat",
        metavar="FILETYPE",
        type=str,
        choices=["png", "pdf", "svg", "eps"],
        default="png",
        help="Image format type. If given, this option "
        "overrides the image format based on the plotFile "
        "ending. (Default: %(default)s)",
    )

    return parser


## Tools: scBulkCoverage, scCountReads, scFilterStats, scJSD
def bamOptions(args=None, suppress_args=None, default_opts=None):
    """
    Arguments for tools that operate on BAM files
    """

    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group("BAM processing options")

    group.add_argument(
        "--cellTag",
        "-ct",
        metavar="STR",
        help="Name of the BAM tag from which to extract barcodes.",
        type=str,
        default="BC",
    )

    group.add_argument(
        "--groupTag",
        "-gt",
        metavar="STR",
        help=show_or_hide(
            "In case of a groupped BAM file, such as the one containing Read Group (`RG`) or Sample (`SM`) tag,"
            "it is possible to process group the reads using the provided --groupTag argument. NOTE: In case of such input, "
            "please ensure that the --labels argument indicates the expected group labels contained in the BAM files. "
            "The --groupTag along with the --cellTag is then used to identify unique samples (cells) from the input.",
            "groupTag",
            suppress_args,
        ),
        type=str,
        default=None,
    )

    group.add_argument(
        "--numberOfProcessors",
        "-p",
        help='Number of processors to use. Type "max/2" to '
        'use half the maximum number of processors or "max" '
        "to use all available processors. (Default: %(default)s)",
        metavar="INT",
        type=numberOfProcessors,
        default=1,
        required=False,
    )

    group.add_argument(
        "--labels",
        "-l",
        metavar="sample1 sample2",
        help=show_or_hide(
            "User defined labels instead of default labels from "
            "file names. Multiple labels have to be separated by a space, e.g. "
            "--labels sample1 sample2 sample3",
            "labels",
            suppress_args,
        ),
        nargs="+",
    )

    group.add_argument(
        "--smartLabels",
        action="store_true",
        help=show_or_hide(
            "Instead of manually specifying labels for the input "
            "BAM files, this causes sincei to use the file name "
            "after removing the path and extension.",
            "smartLabels",
            suppress_args,
        ),
    )

    group.add_argument(
        "--region",
        "-r",
        help=show_or_hide(
            "Region of the genome to limit the operation "
            "to - this is useful when testing parameters to "
            "reduce the computing time. The format is "
            "chr:start:end, for example --region chr10 or --region chr10:456700:891000.",
            "region",
            suppress_args,
        ),
        metavar="CHR:START:END",
        required=False,
        type=genomicRegion,
    )

    group.add_argument(
        "--blackListFileName",
        "-bl",
        help=show_or_hide(
            "A BED or GTF file containing regions that should be excluded from all analyses. Currently this works by rejecting genomic chunks that happen to overlap an entry. Consequently, for BAM files, if a read partially overlaps a blacklisted region or a fragment spans over it, then the read/fragment might still be considered. Please note that you should adjust the effective genome size, if relevant.",
            "blackListFileName",
            suppress_args,
        ),
        metavar="BED file",
        nargs="+",
        required=False,
    )

    group.add_argument(
        "--binSize",
        "-bs",
        help=show_or_hide(
            "Size of the bins, in bases, to calculate coverage (Default: %(default)s)",
            "binSize",
            suppress_args,
        ),
        metavar="INT bp",
        type=int,
        default=get_default("binSize", default_opts),
    )

    group.add_argument(
        "--distanceBetweenBins",
        metavar="INT",
        help=show_or_hide(
            "The gap distance between bins during counting. "
            "Larger numbers can be used to sample the genome for input files with high coverage "
            "while smaller values are useful for lower coverage data. Note that if you specify a value that "
            "results in too few (<1000) reads sampled, the value will be decreased. (Default: %(default)s)",
            "distanceBetweenBins",
            suppress_args,
        ),
        default=get_default("distanceBetweenBins", default_opts),
        type=int,
    )

    return parser


## Tools: scBulkCoverage, scCountReads
def filterOptions(args=None, suppress_args=None):
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group("Read Filtering Options")

    group.add_argument(
        "--duplicateFilter",
        help=show_or_hide(
            "How to filter for duplicates? Different combinations (using start/end/umi) are possible. "
            "Read start position and read barcode are always considered. Default (None) would consider all reads. "
            "Note that in case of paired end data, both reads in the fragment are considered (and kept). So if you wish "
            "to keep only read1, combine this option with samFlagInclude ",
            "duplicateFilter",
            suppress_args,
        ),
        type=str,
        choices=["start_bc", "start_bc_umi", "start_end_bc", "start_end_bc_umi"],
        default=None,
    )

    group.add_argument(
        "--motifFilter",
        "-m",
        metavar="STR",
        help=show_or_hide(
            "Check whether a given motif is present in the read and the corresponding reference genome. "
            "This option checks for the motif at the 5-end of the read and at the 5-overhang in the genome, "
            "which is useful in identifying reads properly cut by a restriction-enzyme or MNAse. "
            'For example, if you want to search for an "A" at the 5\'-end of the read and "TA" at 5\'-overhang, '
            "use \"-m 'A,TA'\". Reads not containing the given motif are filtered out. ",
            "motifFilter",
            suppress_args,
        ),
        type=str,
        nargs="+",
        default=None,
    )

    group.add_argument(
        "--genome2bit",
        "-g",
        metavar="STR",
        help=show_or_hide(
            "If --motifFilter is provided, please also provide the genome sequence (in 2bit format). ",
            "genome2bit",
            suppress_args,
        ),
        type=str,
        default=None,
    )

    group.add_argument(
        "--GCcontentFilter",
        "-gc",
        metavar="STR",
        help=show_or_hide(
            "Check whether the GC content of the read falls within the provided range "
            "Input format must be '<low>,<high>' , where <low> is the lower bound and <high> is the "
            "upper bound in the fraction of GC (eg. '0.1,0.9' ). "
            "If the GC content of the reads fall outside the range, they are filtered out. ",
            "GCcontentFilter",
            suppress_args,
        ),
        type=str,
        default=None,
    )

    group.add_argument(
        "--minAlignedFraction",
        help=show_or_hide(
            "Minimum fraction of the reads which should be aligned to be counted. This includes "
            "mismatches tolerated by the aligners, but excludes InDels/Clippings (Default: %(default)s)",
            "minAlignedFraction",
            suppress_args,
        ),
        metavar="FLOAT",
        default=None,
        type=float,
        required=False,
    )

    return parser


## Tools: scBulkCoverage, scCountReads
def readOptions(args=None, suppress_args=None):
    """Common arguments related to BAM files and the interpretation
    of the read coverage
    """
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group("Read Processing Options")

    group.add_argument(
        "--minMappingQuality",
        metavar="INT",
        help=show_or_hide(
            "If set, only reads that have a mapping " "quality score of at least this are considered.",
            "minMappingQuality",
            suppress_args,
        ),
        type=int,
    )

    group.add_argument(
        "--samFlagInclude",
        help=show_or_hide(
            "Include reads based on the SAM flag. For example, "
            "to get only reads that are the first mate, use a flag of 64. "
            "This is useful to count properly paired reads only once, "
            "as otherwise the second mate will be also considered for the "
            "coverage. (Default: %(default)s)",
            "samFlagInclude",
            suppress_args,
        ),
        metavar="INT",
        default=None,
        type=int,
        required=False,
    )

    group.add_argument(
        "--samFlagExclude",
        help=show_or_hide(
            "Exclude reads based on the SAM flag. For example, "
            "to get only reads that map to the forward strand, use "
            "--samFlagExclude 16, where 16 is the SAM flag for reads "
            "that map to the reverse strand. (Default: %(default)s)",
            "samFlagExclude",
            suppress_args,
        ),
        metavar="INT",
        default=None,
        type=int,
        required=False,
    )

    group.add_argument(
        "--minFragmentLength",
        help=show_or_hide(
            "The minimum fragment length needed for read/pair "
            "inclusion. This option is primarily useful "
            "in ATACseq experiments, for filtering mono- or "
            "di-nucleosome fragments. (Default: %(default)s)",
            "minFragmentLength",
            suppress_args,
        ),
        metavar="INT",
        default=0,
        type=int,
        required=False,
    )

    group.add_argument(
        "--maxFragmentLength",
        help=show_or_hide(
            "The maximum fragment length needed for read/pair " "inclusion. (Default: %(default)s)",
            "maxFragmentLength",
            suppress_args,
        ),
        metavar="INT",
        default=0,
        type=int,
        required=False,
    )

    group.add_argument(
        "--filterRNAstrand",
        help=show_or_hide(
            "Selects RNA-seq reads (single-end or paired-end) originating from genes "
            "on the given strand. This option assumes a standard dUTP-based library "
            "preparation (that is, --filterRNAstrand=forward keeps minus-strand reads, "
            "which originally came from genes on the forward strand using a dUTP-based "
            "method). Consider using --samExcludeFlag instead for filtering by strand in "
            "other contexts.",
            "filterRNAstrand",
            suppress_args,
        ),
        choices=["forward", "reverse"],
        default=None,
    )

    group.add_argument(
        "--extendReads",
        "-e",
        help=show_or_hide(
            "This parameter allows the extension of reads to "
            "fragment size. If set, each read is "
            "extended, without exception.\n"
            "*NOTE*: This feature is generally NOT recommended for "
            "spliced-read data, such as RNA-seq, as it would "
            "extend reads over skipped regions.\n"
            "*Single-end*: Requires a user specified value for the "
            "final fragment length. Reads that already exceed this "
            "fragment length will not be extended.\n"
            "*Paired-end*: Reads with mates are always extended to "
            "match the fragment size defined by the two read mates. "
            "Unmated reads, mate reads that map too far apart "
            "(>4x fragment length) or even map to different "
            "chromosomes are treated like single-end reads. The input "
            "of a fragment length value is optional. If "
            "no value is specified, it is estimated from the "
            "data (mean of the fragment size of all mate reads).\n",
            "extendReads",
            suppress_args,
        ),
        type=int,
        nargs="?",
        const=True,
        default=False,
        metavar="INT bp",
    )

    group.add_argument(
        "--centerReads",
        help=show_or_hide(
            "By adding this option, reads are centered with "
            "respect to the fragment length. For paired-end data, "
            "the read is centered at the fragment length defined "
            "by the two ends of the fragment. For single-end data, the "
            "given fragment length is used. This option is "
            "useful to get a sharper signal around enriched regions.",
            "centerReads",
            suppress_args,
        ),
        action="store_true",
    )

    return parser


## helper functions
def show_or_hide(help, name, arglist):
    if arglist and name in arglist:
        return argparse.SUPPRESS
    else:
        return help


def get_default(name, argdict):
    if argdict and name in argdict.keys():
        return argdict[name]
    else:
        return None


def process_args(args=None):
    if not args.labels:
        if args.smartLabels:
            args.labels = smartLabels(args.bamfiles)
        else:
            args.labels = args.bamfiles

    if len(args.bamfiles) != len(args.labels):
        sys.exit("The number of labels does not match the number of BAM files.")

    return args


def genomicRegion(string):
    # remove whitespaces using split,join trick
    region = "".join(string.split())
    if region == "":
        return None
    # remove undesired characters that may be present and
    # replace - by :
    # N.B., the syntax for translate() differs between python 2 and 3
    try:
        region = region.translate(None, ",;|!{}()").replace("-", ":")
    except:
        region = region.translate({ord(i): None for i in ",;|!{}()"})
    if len(region) == 0:
        raise argparse.ArgumentTypeError("{} is not a valid region".format(string))
    return region


def numberOfProcessors(string):
    import multiprocessing

    availProc = multiprocessing.cpu_count()

    if string == "max/2":  # default case
        # by default half of the available processors are used
        numberOfProcessors = int(availProc * 0.5)
    elif string == "max":
        # use all available processors
        numberOfProcessors = availProc
    else:
        try:
            numberOfProcessors = int(string)
        except ValueError:
            raise argparse.ArgumentTypeError("{} is not a valid number of processors".format(string))

        except Exception as e:
            raise argparse.ArgumentTypeError(
                "the given value {} is not valid. "
                "Error message: {}\nThe number of "
                "available processors in your "
                "computer is {}.".format(string, e, availProc)
            )

        if numberOfProcessors > availProc:
            numberOfProcessors = availProc

    return numberOfProcessors


def smartLabel(label):
    """
    Remove the path name and the last extension from the file name
    /pth/to/file.name.bam -> file.name
    """
    lab = os.path.splitext(os.path.basename(label))[0]
    if lab == "":
        # Maybe we have a dot file?
        lab = os.path.basename(label)
    return lab


def smartLabels(labels):
    smrt = [smartLabel(x) for x in labels]
    if len(smrt) != len(set(smrt)):
        print(
            "Labels inferred from file names are not unique. "
            "Please be aware that in case of overlapping barcodes the counts will be merged."
        )
    return smrt


def validateInputs(args, process_barcodes=True):
    """
    Ensure that right input is provided from argparse
    """

    ## Labels
    if args.groupTag:
        # in case of --groupTag, use args.labels as groups
        if len(args.bamfiles) > 1:
            sys.stderr.write("Only a single BAM file is allowed when --groupTag is specified.")
            exit(0)
        if not args.labels:
            sys.stderr.write("Please indicate the sample groups to be processed from the BAM file with --labels")
            exit(0)
    else:
        if args.labels and len(args.bamfiles) != len(args.labels):
            sys.stderr.write(
                "The number of labels does not match the number of bam files. "
                "This is only allowed if a single BAM file is provided and --groupTag is specified."
            )
            exit(0)
        if not args.labels:
            args.labels = smartLabels(args.bamfiles)

    ## Motif and GC filter
    if args.motifFilter:
        if not args.genome2bit:
            sys.stderr.write("MotifFilter asked but genome (2bit) file not provided.")
            sys.exit(1)
        else:
            args.motifFilter = [x.strip(" ").split(",") for x in args.motifFilter]

    if args.GCcontentFilter:
        gc = args.GCcontentFilter.strip(" ").split(",")
        args.GCcontentFilter = [float(x) for x in gc]

    ## Barcodes + new labels (if required)
    if process_barcodes:
        with open(args.barcodes, "r") as f:
            barcodes = f.read().splitlines()
        f.close()
        args.barcodes = barcodes
        newlabels = ["{}::{}".format(a, b) for a in args.labels for b in barcodes]
        return args, newlabels
    else:
        return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys

from sincei._version import __version__


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
sincei is a suite of command-line tools developed for a user-friendly analysis of single-cell sequencing data.
Version: {}

Each tool begins with the prefix sc<tool_name>, such as:

 $ scBulkCoverage -b file1.bam -g groupinfo.txt -o coverage

[ Tools for a typical single-cell analysis workflow ] (WIP: work in progress/not available yet)

    scFilterBarcodes        Identify and filter cell barcodes from BAM file (for droplet-based single-cell seq)
    scFilterStats           Produce per-cell statistics after filtering reads by user-defined criteria.
    scCountReads            Counts reads for each barcode on genomic bins or user-defined features.
    scCountQC               Perform quality control and filter the output of scCountReads.
    scCombineCounts         [WIP] Concatenate/merge the counts from different samples/batches or modalities
    scClusterCells          Perform dimensionality reduction and clustering on the output of scCountReads.
    scBulkCoverage          Get pseudo-bulk coverage per group using a user-supplied cell->group mapping (output of scClusterCells).
    scBAMops                Modify a BAM file to group cells (using cell barcodes), or filter/shift mapped reads.
    scFindMarkers           [WIP] Find marker genes per group, given the output of scCountReads and a user-defined group.
    scFeaturePlot           [WIP] Plot the counts for a given feature on a UMAP or on a (IGV-style) genomic-track.

""".format(
            __version__
        ),
    )

    return parser


def process_args(args=None):
    args = parse_arguments().parse_args(args)

    return args


def main(args=None):
    if args is None and len(sys.argv) == 1:
        args = ["--help"]
    process_args(args)



import pandas as pd
import numpy as np
import numpy.testing as nt
import torch
import scipy
from sincei.ExponentialFamily import Bernoulli, Gaussian, Poisson, Beta, SigmoidBeta, Gamma


def testGaussianDistribution():
    # Verifies compliance of Gaussian definition using randomly selected points.
    X = torch.linspace(-5, 5, 500)
    for _ in range(25):
        # random parameter (mean)
        theta = torch.normal(0.0, 1.0, (1, 1))[0]

        # Gaussian PDF generated by our implementation
        gaussian_dist = Gaussian()
        dist_gauss_values = np.array([gaussian_dist.distribution(x, theta).detach().numpy() for x in X]).flatten()

        # Scipy Gaussian distribution
        scipy_gauss_val = scipy.stats.norm.pdf(X, loc=theta)

        # Assert equality
        nt.assert_array_almost_equal(dist_gauss_values, scipy_gauss_val, decimal=3)


def testBernoulliDistribution():
    # Verifies compliance of Bernoulli definition using randomly selected points.
    X = torch.Tensor([0, 1])

    # Span possible values from -30 to 30
    theta_val = torch.linspace(-30, 30, 1000)
    for theta in theta_val:
        bernoulli_dist = Bernoulli()
        dist_bernoulli_values = np.array([bernoulli_dist.distribution(x, theta).detach().numpy() for x in X]).flatten()

        # Compare to theoretical value
        assert np.abs(dist_bernoulli_values[-1] - 1 / (1 + torch.exp(-theta))) < 1e-6


def testPoissonDistribution():
    # Verifies compliance of Poisson definition using randomly selected points.
    X = torch.linspace(0, 100, 101)

    # Span possible values from -30 to 30
    theta_val = torch.linspace(-50, 5, 10)
    for theta in theta_val:
        poisson_dist = Poisson()
        log_dist_poisson_values = np.array(
            [poisson_dist.log_distribution(x, theta).detach().numpy() for x in X]
        ).flatten()

        # Compare to Scipy values
        np.testing.assert_array_almost_equal(
            log_dist_poisson_values, scipy.stats.poisson.logpmf(X, np.exp(theta)), decimal=3
        )


def testBetaDistribution():
    # Verifies compliance of Beta definition using randomly selected points.
    X = torch.linspace(0 + 1e-4, 1 - 1e-4, 100)

    # Possible parameters: scan through them
    theta_val = torch.linspace(0.01, 0.99, 10)
    nu_val = torch.rand(20) * 10
    for nu in nu_val:
        for theta in theta_val:
            # Our definition
            beta_dist = Beta()
            beta_dist.family_params["nu"] = torch.Tensor([nu])
            log_dist_beta_values = np.array(
                [beta_dist.log_distribution(x, theta).detach().numpy() for x in X]
            ).flatten()
            dist_beta_values = np.array([beta_dist.distribution(x, theta).detach().numpy() for x in X]).flatten()

            # Compare to Scipy.
            np.testing.assert_array_almost_equal(
                log_dist_beta_values, scipy.stats.beta.logpdf(X, theta * nu, nu * (1 - theta)), decimal=2
            )


def testSigmoidBetaDistribution():
    # Verifies compliance of SigmoidBeta definition using randomly selected points.
    X = torch.linspace(0 + 1e-4, 1 - 1e-4, 100)
    # Possible parameters
    theta_val = torch.logit(torch.linspace(0.01, 0.99, 10))
    nu_val = torch.rand(20) * 10

    for nu in nu_val:
        for theta in theta_val:
            # Our definition
            beta_dist = SigmoidBeta()
            beta_dist.family_params["nu"] = torch.Tensor([nu])
            log_dist_beta_values = np.array(
                [beta_dist.log_distribution(x, theta).detach().numpy() for x in X]
            ).flatten()
            dist_beta_values = np.array([beta_dist.distribution(x, theta).detach().numpy() for x in X]).flatten()

            # Comparison to Scipy.
            np.testing.assert_array_almost_equal(
                log_dist_beta_values,
                scipy.stats.beta.logpdf(X, torch.sigmoid(theta) * nu, nu * (1 - torch.sigmoid(theta))),
                decimal=2,
            )


def testGammaDistribution():
    # Verifies compliance of Gamma definition using randomly selected points.
    X = torch.logspace(-8, 5, 1000)
    theta_val = torch.logspace(-5, 3, 50)

    # Possible parameters
    nu_val = torch.rand(20) * 10

    for nu in nu_val:
        for theta in theta_val:
            # Our definition
            gamma_dist = Gamma()
            gamma_dist.family_params["nu"] = torch.Tensor([nu])
            log_dist_gamma_values = np.array(
                [gamma_dist.distribution(torch.Tensor(x), torch.Tensor([theta])).detach().numpy() for x in X]
            ).flatten()

            # Comparison to Scipy.
            np.testing.assert_array_almost_equal(
                scipy.stats.gamma.pdf(X, a=theta + 1, loc=0, scale=1 / nu), log_dist_gamma_values, decimal=2
            )



from itertools import compress
from deeptools.utilities import getTLen
import numpy as np
import sys


def checkBAMtag(bam, name, tag):
    bctag = [read.has_tag(tag) for read in bam.head(1000)]
    if not any(bctag):
        sys.stderr.write("WARNING: Input file {} seems to lack the tag {}. Output might be empty. \n".format(name, tag))
    return None


def checkMotifs(read, chrom, genome, readMotif, refMotif):
    """
    Check whether a given motif is present in the read and the corresponding reference genome.
    For example, in MNAse (scChIC-seq) data, we expect the reads to have an 'A' at the 5'-end,
    while the genome has a 'TA' over hang (where the 'A' aligns with 'A' in the forward read),
    like this below.

    Forwards aligned read: read has 'A', upstream has T
    R1 ........A------->
    ----------TA------------\ Ref (+)

    Rev aligned read: read has 'T', downstream has A

    <-------T....... R1
    --------TA------------\ Ref (+)

    This function can look for any arbitrary motif in read and corresponding genome, but in the
    same orientation as described above.

    :return: bool

    >>> import pysam
    >>> import os
    >>> from scDeepTools.scReadCounter import CountReadsPerBin as cr
    >>> root = os.path.dirname(os.path.abspath(__file__)) + "/test/test_data/"
    >>> bam = pysam.AlignmentFile("{}/test_TA_filtering.bam".format(root))
    >>> iter = bam.fetch()
    >>> read = next(iter)
    >>> cr.checkMotifs(read, 'A', 'TA') # for valid scChIC read
    True
    >>> read = next(iter)
    >>> cr.checkMotifs(read, 'A', 'TA') # for invalid scChIC read
    False
    """
    # get read and ref motif pos
    read_motif = read.get_forward_sequence()[0 : len(readMotif)]
    ref_motifLen = len(refMotif) - 1

    if read.is_reverse:
        # for reverse reads ref motif begins at read-end and ends downstream
        endpos = read.reference_end + ref_motifLen
        if endpos > genome.chroms()[chrom]:
            endpos = read.reference_end  # fail without error
        ref_motif = genome.sequence(chrom, read.reference_end - 1, endpos)
    else:
        # for forward reads ref motif begins upstream and ends at read-start
        startpos = read.reference_start - ref_motifLen
        if startpos < 0:
            startpos = read.reference_start  # fail without error
        ref_motif = genome.sequence(chrom, startpos, read.reference_start + 1)

    if read_motif == readMotif and ref_motif == refMotif:
        return True
    else:
        return False


def checkGCcontent(read, lowFilter, highFilter, returnGC=False):
    r"""Checks if the GC content of the read is within the given range

    Parameters
    ----------
    read : pysam.AlignedSegment
        A pysam AlignedSegment object
    lowFilter : float
        Minimum GC content
    highFilter : float
        Maximum GC content
    returnGC : bool
        If true, return the GC content of the read

    Returns
    -------
    bool
        True if the GC content of the read is within the given range

    Examples
    --------
    >>> test = Tester()
    >>> read = test.bamFile1.fetch().next()
    >>> checkGCcontent(read, 0.3, 0.7)
    True
    """
    seq = read.get_forward_sequence()
    total_bases = len(seq)
    gc_bases = len([x for x in seq if x == "C" or x == "G"])
    gc_frac = float(gc_bases) / total_bases
    if returnGC:
        return gc_frac
    else:
        if gc_frac >= lowFilter and gc_frac <= highFilter:
            return True
        else:
            return False


def checkAlignedFraction(read, lowFilter):
    """
    Check whether the fraction of read length that aligns to the reference is higher than
    the given threshold. Aligned fraction includes the max allowed mismatches tolerated by
    the aligner, and excludes InDels and Clippings.

    Return: Bool
    """
    cig = read.cigartuples
    tot = read.infer_read_length()
    matchPos = [i[0] == 0 for i in cig]
    matchSum = sum([i[1] for i in list(compress(cig, matchPos))])
    if matchSum / tot >= lowFilter:
        return True
    else:
        return False


def colorPicker(name):
    r"""
    This function returns a list of colors for plotting.

    Parameters
    ----------
    name : str
        The name of the color palette to use.

    Returns
    -------
    list
        A list of colors.

    Examples
    --------
    >>> colorPicker('twentyfive')
    ['#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00', '#ffff33', '#a65628', '#f781bf', '#999999']

    >>> colorPicker('colorblind')
    ['#0072B2', '#009E73', '#D55E00', '#CC79A7', '#F0E442', '#56B4E9', '#E69F00', '#000000']
    """

    colors = {
        "twentyfive": [
            "#dd4e34",
            "#78d545",
            "#b047dc",
            "#d6d94d",
            "#5d48c6",
            "#74d88b",
            "#c944af",
            "#8a9739",
            "#542c75",
            "#d3953c",
            "#607dc7",
            "#487f46",
            "#d04774",
            "#7bd2c8",
            "#6b2737",
            "#cfcb9a",
            "#332a42",
            "#d7928a",
            "#343d25",
            "#cc8ad3",
            "#7b6a43",
            "#b5bad9",
            "#99472b",
            "#4e8290",
            "#936987",
        ],
        "colorblind": [
            "#ffaec0",
            "#cd7600",
            "#893558",
            "#195f37",
            "#da71f9",
            "#a2d391",
            "#881e9b",
            "#b9d05f",
            "#524d7e",
            "#f2bf4b",
            "#01d6c2",
            "#f54040",
            "#0097fb",
            "#756400",
            "#4b44aa",
            "#f0bd79",
            "#a1008b",
            "#6d4d02",
            "#ff9afc",
            "#01df8f",
            "#e2b8ed",
            "#6e9d00",
            "#f4177b",
            "#01b65f",
            "#9b2532",
        ],
    }

    return colors[name]


def getDupFilterTuple(read, bc, filterArg):
    r"""
    Returns a tuple with the information needed to filter duplicates, based on read and filter type.
    The tuple is composed of the barcode, the umi, the start and end positions
    and the chromosome name.

    Parameters
    ----------
    read : pysam.AlignedSegment
        A pysam.AlignedSegment object.
    bc : str
        The barcode.
    filter : str
        A string with the type of filter to use.

    Returns
    -------
    tuple
        A tuple with the information needed to filter duplicates.
        The tuple is composed of the barcode, the umi, the start and end positions
        and the chromosome name.

    Examples
    --------
    >>> test = Tester()
    >>> read = test.bamFile1.fetch().next()
    >>> getDupFilterTuple(read, 'ATCG', 'end_umi')
    ('ATCG', 'ATCG', None, None, 0, False)
    """

    tLenDup = getTLen(read, notAbs=True)
    filt = filterArg.split("_")

    # get fragment start and end for that read
    if tLenDup >= 0:
        s = read.pos
        e = s + tLenDup
    else:
        s = read.pnext
        e = s - tLenDup

    if "end" not in filt:
        # ignore read/fragment end and mate information
        mate_refid = read.reference_id
        if read.is_reverse:
            s = None
        else:
            e = None
    else:
        # use mate info, reset fragment end to mate pos if read is chimeric
        if read.reference_id != read.next_reference_id:
            e = read.pnext
        mate_refid = read.next_reference_id

    # get UMI if asked
    if "umi" in filt:
        try:
            umi = read.get_tag("RX")
        except KeyError:
            sys.stderr.write("UMI tag (RX) absent, skipping UMI check")
            umi = None
    else:
        umi = None

    tup = (bc, umi, s, e, mate_refid, read.is_reverse)
    return tup


def gini(i, X):
    r"""Computes the Gini coefficient for each row of a sparse matrix (Obs*Var).

    Parameters
    ----------
    i : int
        row index
    X : numpy array
        matrix

    Returns
    -------
    float
        Gini coefficient for the given row

    Examples
    --------
    >>> X = np.matrix([[1,2,3,4],[5,6,7,8],[9,10,11,12]])
    >>> gini(0, X)
    0.0
    >>> gini(1, X)
    0.0
    >>> gini(2, X)
    0.0
    """

    # based on bottom eq:
    # http://www.statsdirect.com/help/generatedimages/equations/equation154.svg
    # from:
    # http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm
    # All values are treated equally, arrays must be 1d:
    try:
        array = X[i, :].toarray().flatten()  # get all bins from i'th cell
    except AttributeError:
        array = X[i, :].flatten()
    array = array[array.nonzero()]

    if array.shape[0] <= 1:
        return np.nan
    else:
        array = np.sort(array)
        # Index per array element:
        index = np.arange(1, array.shape[0] + 1)
        # Number of array elements:
        n = array.shape[0]
        # Gini coefficient:
        return (np.sum((2 * index - n - 1) * array)) / (n * np.sum(array))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import pysam
import os
import sys
import pandas as pd
import py2bit
from deeptools import parserCommon
from deeptools.bamHandler import openBam
from deeptools.mapReduce import mapReduce
from deeptools.utilities import getTLen, smartLabels, getTempFileName

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei import ParserCommon
from sincei.Utilities import checkMotifs, checkGCcontent, getDupFilterTuple
from sincei._version import __version__

## UPDATE: add group tag to BAM file based on a 2-columns mapping file (barcode -> group)


def parseArguments():
    internalParser = get_args()
    ioParser = ParserCommon.inputOutputOptions(opts=["bamfile", "groupInfo", "outFile"])
    bamParser = ParserCommon.bamOptions(suppress_args=["binSize", "distanceBetweenBins"])
    filterParser = ParserCommon.filterOptions()
    readParser = ParserCommon.readOptions(suppress_args=["extendReads", "centerReads"])
    otherParser = ParserCommon.otherOptions()
    parser = argparse.ArgumentParser(
        parents=[
            ioParser,
            internalParser,
            bamParser,
            filterParser,
            readParser,
            otherParser,
        ],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="This tool filters alignments in a BAM/CRAM file according the the specified parameters. It can optionally output to BEDPE format.",
        usage="Example usage: scBAMops -b sample1.bam -o sample1.filtered.bam --minMappingQuality 10 --filterMetrics log.txt",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)

    mod = parser.add_argument_group("Read modifiction arguments")
    mod.add_argument(
        "--shift",
        nargs="+",
        type=int,
        help='Shift the left and right end of a read (for BAM files) or a fragment (for BED files). A positive value shift an end to the right (on the + strand) and a negative value shifts a fragment to the left. Either 2 or 4 integers can be provided. For example, "2 -3" will shift the left-most fragment end two bases to the right and the right-most end 3 bases to the left. If 4 integers are provided, then the first and last two refer to fragments whose read 1 is on the left or right, respectively. Consequently, it is possible to take strand into consideration for strand-specific protocols. A fragment whose length falls below 1 due to shifting will not be written to the output. See the online documentation for graphical examples. Note that non-properly-paired reads will be filtered.',
    )

    mod.add_argument(
        "--ATACshift",
        action="store_true",
        help="Shift the produced BAM file or BEDPE regions as commonly done for ATAC-seq. This is equivalent to --shift 4 -5 5 -4.",
    )

    output = parser.add_argument_group("Optional output arguments")
    output.add_argument(
        "--BED",
        action="store_true",
        help="Instead of producing BAM files, write output in BEDPE format (as defined by MACS2). Note that only reads/fragments passing filtering criterion are written in BEDPE format.",
    )

    output.add_argument(
        "--filterMetrics",
        metavar="FILE.log",
        help="The number of entries in total and filtered are saved to this file",
    )

    output.add_argument(
        "--filteredOutReads",
        metavar="filtered.bam",
        help="If desired, all reads NOT passing the filtering criteria can be written to this file.",
    )

    output.add_argument(
        "--outTagName",
        metavar="STR",
        type=str,
        help="In case where you want to group the BAM files based on user-defined --groupInfo, specify the output BAM tag which would contain the group name.",
        default=None,
    )

    return parser


def shiftRead(b, chromDict, args):
    if not b.is_proper_pair:
        return None
    tLen = getTLen(b, notAbs=True)
    start = b.pos
    end = start + b.query_alignment_end
    if b.is_reverse and not b.is_read2:
        end -= args.shift[2]
        deltaTLen = args.shift[3] - args.shift[2]
    elif b.is_reverse and b.is_read2:
        end += args.shift[1]
        deltaTLen = args.shift[1] - args.shift[0]
    elif not b.is_reverse and not b.is_read2:
        start += args.shift[0]
        deltaTLen = args.shift[1] - args.shift[0]
    else:
        start -= args.shift[3]
        deltaTLen = args.shift[3] - args.shift[2]

    # Sanity check
    if end - start < 1:
        if b.is_reverse:
            start = end - 1
        else:
            end = start + 1
    if start < 0:
        start = 0
    if end > chromDict[b.reference_name]:
        end = chromDict[b.reference_name]
    if end - start < 1:
        return None

    # create a new read
    b2 = pysam.AlignedSegment()
    b2.query_name = b.query_name
    b2.flag = b.flag
    b2.reference_id = b.reference_id
    b2.reference_start = start
    b2.mapping_quality = b.mapping_quality
    b2.cigar = ((0, end - start),)  # Returned cigar is only matches
    if tLen < 0:
        b2.template_length = tLen - deltaTLen
    else:
        b2.template_length = tLen + deltaTLen
    b2.next_reference_id = b.next_reference_id
    b2.next_reference_start = b.next_reference_start
    if b.is_proper_pair:
        if b2.is_read2 and b2.is_reverse:
            b2.next_reference_start += args.shift[0]
        elif not b2.is_read2 and b2.is_reverse:
            b2.next_reference_start -= args.shift[3]

    return b2


def filterWorker(arglist):
    chrom, start, end, args, chromDict = arglist
    fh = openBam(args.bamfile)
    # open 2 bit if needed
    if args.genome2bit:
        twoBitGenome = py2bit.open(args.genome2bit, True)
        if chrom not in twoBitGenome.chroms().keys():
            raise NameError("chromosome {} not found in 2bit file".format(chrom))

    mode = "wbu"
    oname = getTempFileName(suffix=".bam")
    ofh = pysam.AlignmentFile(oname, mode=mode, template=fh)

    if args.filteredOutReads:
        onameFiltered = getTempFileName(suffix=".bam")
        ofiltered = pysam.AlignmentFile(onameFiltered, mode=mode, template=fh)
    else:
        onameFiltered = None
        ofiltered = None

    prev_pos = set()
    lpos = None

    nFiltered = 0
    total = 0
    for read in fh.fetch(chrom, start, end):
        if read.pos < start:
            # ensure that we never double count (in case distanceBetweenBins == 0)
            continue
        total += 1

        try:
            bc = read.get_tag(args.cellTag)
        except KeyError:
            nFiltered += 1
            continue
        if isinstance(args.groupInfo, pd.DataFrame):
            smpl = read.get_tag(args.groupTag)
            try:
                tag_to_add = args.groupInfo.loc[
                    (args.groupInfo["sample"] == smpl) & (args.groupInfo["barcode"] == bc),
                    "cluster",
                ].values.item()
                read.set_tag(args.outTagName, tag_to_add, value_type="Z", replace=True)
            except:
                if args.verbose:
                    sys.stderr.write("Encountered read tags not in groupInfo file, skipped..")
                nFiltered += 1
                if ofiltered:
                    ofiltered.write(read)
                continue

        ## -- begin filtering -- ##
        if read.flag & 4:
            # Ignore unmapped reads, they were counted already
            nFiltered += 1
            if ofiltered:
                ofiltered.write(read)
            continue

        if args.minMappingQuality and read.mapq < args.minMappingQuality:
            nFiltered += 1
            if ofiltered:
                ofiltered.write(read)
            continue

        if args.samFlagInclude and read.flag & args.samFlagInclude != args.samFlagInclude:
            nFiltered += 1
            if ofiltered:
                ofiltered.write(read)
            continue
        if args.samFlagExclude and read.flag & args.samFlagExclude != 0:
            nFiltered += 1
            if ofiltered:
                ofiltered.write(read)
            continue

        tLen = getTLen(read)
        if args.minFragmentLength > 0 and tLen < args.minFragmentLength:
            nFiltered += 1
            if ofiltered:
                ofiltered.write(read)
            continue
        if args.maxFragmentLength > 0 and tLen > args.maxFragmentLength:
            nFiltered += 1
            if ofiltered:
                ofiltered.write(read)
            continue

        if args.duplicateFilter:
            tup = getDupFilterTuple(read, bc, args.duplicateFilter)
            if lpos is not None and lpos == read.reference_start and tup in prev_pos:
                nFiltered += 1
                if ofiltered:
                    ofiltered.write(read)
                continue
            if lpos != read.reference_start:
                prev_pos.clear()
            lpos = read.reference_start
            prev_pos.add(tup)

        # remove reads with low/high GC content
        if args.GCcontentFilter:
            if not checkGCcontent(read, args.GCcontentFilter[0], args.GCcontentFilter[1]):
                nFiltered += 1
                if ofiltered:
                    ofiltered.write(read)
                continue

        # remove reads that don't pass the motif filter
        if args.motifFilter:
            if not checkMotifs(read, chrom, twoBitGenome, args.motifFilter[0], args.motifFilter[1]):
                nFiltered += 1
                if ofiltered:
                    ofiltered.write(read)
                continue

        # filterRNAstrand
        if args.filterRNAstrand:
            if read.is_paired:
                if args.filterRNAstrand == "forward":
                    if read.flag & 144 == 128 or read.flag & 96 == 64:
                        pass
                    else:
                        nFiltered += 1
                        if ofiltered:
                            ofiltered.write(read)
                        continue
                elif args.filterRNAstrand == "reverse":
                    if read.flag & 144 == 144 or read.flag & 96 == 96:
                        pass
                    else:
                        nFiltered += 1
                        if ofiltered:
                            ofiltered.write(read)
                        continue
            else:
                if args.filterRNAstrand == "forward":
                    if read.flag & 16 == 16:
                        pass
                    else:
                        nFiltered += 1
                        if ofiltered:
                            ofiltered.write(read)
                        continue
                elif args.filterRNAstrand == "reverse":
                    if read.flag & 16 == 0:
                        pass
                    else:
                        nFiltered += 1
                        if ofiltered:
                            ofiltered.write(read)
                        continue

        if args.shift:
            read = shiftRead(read, chromDict, args)
            if not read:
                continue

        # Read survived filtering
        ofh.write(read)

    # The results from the workers will get sorted, so get the TID
    tid = fh.get_tid(chrom)

    ofh.close()
    if ofiltered:
        ofiltered.close()
    fh.close()
    if args.genome2bit:
        twoBitGenome.close()

    return tid, start, total, nFiltered, oname, onameFiltered


def convertBED(oname, tmpFiles, chromDict):
    """
    Stores results in BEDPE format, which is:
    chromosome	frag_leftend	frag_rightend

    The fragment ends can be shifted
    """
    ofile = open(oname, "w")
    for tmpFile in tmpFiles:
        fh = pysam.AlignmentFile(tmpFile)

        for b in fh.fetch(until_eof=True):
            tLen = getTLen(b, notAbs=True)
            if tLen > 0:
                start = b.pos
                end = start + tLen
                if end > chromDict[b.reference_name]:
                    end = chromDict[b.reference_name]
                if end - start < 1:
                    continue
                ofile.write("{}\t{}\t{}\n".format(b.reference_name, start, end))
        fh.close()
        os.unlink(tmpFile)
    ofile.close()


def main(args=None):
    args = parseArguments().parse_args(args)
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")
    # grouping and tagging the output alignments
    if args.groupInfo:
        if not args.outTagName:
            sys.stderr.write(
                "--groupInfo provided without --outTagName. By default, groups will be added to the 'RG' tag in the output BAM file."
            )
            args.outTagName = "RG"

        df = pd.read_csv(args.groupInfo, sep="\t", index_col=None, comment="#")
        if len(df.columns) == 3:
            df.columns = ["sample", "barcode", "cluster"]
            df.index = df[["sample", "barcode"]].apply(lambda x: "::".join(x), axis=1)
            args.groupInfo = df
        else:
            sys.stderr.write(
                "Error: The input to --groupInfo must be a 3-column tsv file with <sample> <barcode> and <group>"
            )

    # shifting the output alignments
    if args.shift:
        if len(args.shift) not in [2, 4]:
            sys.exit("The --shift option can accept either 2 or 4 values only.")
        if len(args.shift) == 2:
            args.shift.extend([-args.shift[1], -args.shift[0]])
    elif args.ATACshift:
        args.shift = [4, -5, 5, -4]

    bam, mapped, unmapped, stats = openBam(args.bamfile, returnStats=True, nThreads=args.numberOfProcessors)
    total = mapped + unmapped
    chrom_sizes = [(x, y) for x, y in zip(bam.references, bam.lengths)]
    chromDict = {x: y for x, y in zip(bam.references, bam.lengths)}

    # if motifFilter is given, convert the input to list
    if args.motifFilter:
        if not args.genome2bit:
            print("MotifFilter asked but genome (2bit) file not provided.")
            sys.exit(1)
        else:
            args.motifFilter = args.motifFilter.strip(" ").split(",")

    if args.GCcontentFilter:
        gc = args.GCcontentFilter.strip(" ").split(",")
        args.GCcontentFilter = [float(x) for x in gc]

    # Filter, writing the results to a bunch of temporary files
    res = mapReduce(
        [args, chromDict],
        filterWorker,
        chrom_sizes,
        region=args.region,
        blackListFileName=args.blackListFileName,
        numberOfProcessors=args.numberOfProcessors,
        verbose=args.verbose,
    )

    res = sorted(res)  # The temp files are now in order for concatenation
    nFiltered = sum([x[3] for x in res])
    totalSeen = sum([x[2] for x in res])  # The * contig isn't queried

    tmpFiles = [x[4] for x in res]
    if not args.BED:
        arguments = ["-o", args.outFile]
        arguments.extend(tmpFiles)  # [..., *someList] isn't available in python 2.7
        pysam.samtools.cat(*arguments)
        for tmpFile in tmpFiles:
            os.unlink(tmpFile)
    else:
        convertBED(args.outFile, tmpFiles, chromDict)

    if args.filteredOutReads:
        tmpFiles = [x[5] for x in res]
        if not args.BED:
            arguments = ["-o", args.filteredOutReads]
            arguments.extend(tmpFiles)  # [..., *someList] isn't available in python 2.7
            pysam.samtools.cat(*arguments)
            for tmpFile in tmpFiles:
                os.unlink(tmpFile)
        else:
            convertBED(args.outFile, tmpFiles, chromDict, args)

    if args.filterMetrics:
        sampleName = args.bamfile
        if args.label:
            sampleName = args.label
        if args.smartLabels:
            sampleName = smartLabels([args.bamfile])[0]

        of = open(args.filterMetrics, "w")
        of.write("#bamFilterReads --filterMetrics\n")
        of.write("#File\tReads Remaining\tTotal Initial Reads\n")
        of.write("{}\t{}\t{}\n".format(sampleName, totalSeen - nFiltered, total))
        of.close()

    return 0



import os

from deeptools import bamHandler, utilities
from deeptoolsintervals import GTF

import py2bit

from sincei.Utilities import (
    checkMotifs,
    checkGCcontent,
    checkAlignedFraction,
    getDupFilterTuple,
)


def getStats_worker(arglist):
    r"""Computes statistics for each read in a bam file

    This function computes statistics for each read in a bam file.

    Parameters
    ----------
    bamfiles : list
        List containing the names of indexed bam files. E.g. ['file1.bam', 'file2.bam']
    binSize : int
        Length of the window/bin. This value is overruled by ``bedFile`` if present.
    barcodes : list
        list of barcodes to count the reads from.
    numberOfSamples : int
        Total number of samples. The genome is divided into ``numberOfSamples``, each
        with a window/bin length equal to ``binLength``. This value is overruled
        by ``stepSize`` in case such value is present and by ``bedFile`` in which
        case the number of samples and bins are defined in the bed file

    numberOfProcessors : int
        Number of processors to use. Default is 4

    verbose : bool
        Output messages. Default: False

    region : str
        Region to limit the computation in the form chrom:start
    """

    chrom, start, end, args = arglist
    # Fix the bounds
    if end - start > args.binSize and end - start > args.distanceBetweenBins:
        end -= args.distanceBetweenBins
    if end <= start:
        end = start + 1
    ## open genome if needed
    if args.genome2bit:
        twoBitGenome = py2bit.open(args.genome2bit, True)
    ## open blacklist file
    blackList = None
    if args.blackListFileName is not None:
        blackList = GTF(args.blackListFileName)

    o = []
    for fname in args.bamfiles:
        fh = bamHandler.openBam(fname)
        chromUse = utilities.mungeChromosome(chrom, fh.references)
        prev_pos = set()
        lpos = None
        ## initiate a dict with all values to keep per read
        info_list = []

        for read in fh.fetch(chromUse, start, end):
            ## general filtering
            if read.pos < start:
                # ensure that we never double count (in case distanceBetweenBins == 0)
                continue
            if read.flag & 4:
                # Ignore unmapped reads, they were counted already
                continue
            if args.minMappingQuality and read.mapq < args.minMappingQuality:
                continue
            if args.minAlignedFraction:
                if not checkAlignedFraction(read, args.minAlignedFraction):
                    continue
            if blackList and blackList.findOverlaps(
                chrom,
                read.reference_start,
                read.reference_start + read.infer_query_length(always=False) - 1,
            ):
                continue
            if args.motifFilter:
                test = [checkMotifs(read, chrom, twoBitGenome, m[0], m[1]) for m in args.motifFilter]
                # if none given motif found, return true
                if not any(test):
                    continue

            # now collect info
            info = [None for n in range(0, 8)]
            info[0] = chromUse
            if args.barcodes is not None:
                try:
                    bc = read.get_tag(args.cellTag)
                    if bc in args.barcodes:
                        info[1] = True
                    else:
                        info[1] = False
                except KeyError:
                    continue
            else:
                info[1] = False
                bc = None
            info[2] = read.reference_start

            ## Duplicates
            tup = getDupFilterTuple(read, bc, args.duplicateFilter)
            if lpos is not None and lpos == read.reference_start and tup in prev_pos:
                info[3] = True  # read is duplicate
                info[4] = 0
            else:
                info[3] = False
            if lpos != read.reference_start:
                prev_pos.clear()
                ## add distance to last read
                if lpos is None:
                    info[4] = 0
                else:
                    info[4] = abs(float(lpos - read.reference_start))
                ## also add the distance of this read to previous pos

            lpos = read.reference_start
            prev_pos.add(tup)

            info[5] = checkGCcontent(read, args.GCcontentFilter[0], args.GCcontentFilter[1], returnGC=True)
            # filterRNAstrand
            info[6] = read.is_reverse
            if args.getReadID:
                info[7] = read.query_name
            info_list.append(info)
        fh.close()

        # out is an array with row = len(barcode) [384], column = len(stats) [11]
        o.append(info_list)
    return o



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys

from deeptools import bamHandler, utilities
from deeptools.mapReduce import mapReduce
from deeptoolsintervals import GTF

import numpy as np
import py2bit
import pandas as pd

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei.Utilities import (
    checkMotifs,
    checkGCcontent,
    checkBAMtag,
    checkAlignedFraction,
    getDupFilterTuple,
)
from sincei import ParserCommon


def parseArguments():
    filterParser = ParserCommon.filterOptions()

    io_args = ParserCommon.inputOutputOptions(opts=["bamfiles", "barcodes", "outFile"], requiredOpts=["barcodes"])
    bam_args = ParserCommon.bamOptions(
        suppress_args=["region", "groupTag"],
        default_opts={"binSize": 100000, "distanceBetweenBins": 1000000},
    )
    filter_args = ParserCommon.filterOptions()
    read_args = ParserCommon.readOptions(
        suppress_args=[
            "minFragmentLength",
            "maxFragmentLength",
            "extendReads",
            "centerReads",
        ]
    )
    other_args = ParserCommon.otherOptions()

    parser = argparse.ArgumentParser(
        parents=[io_args, bam_args, filter_args, read_args, other_args],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
This tool estimates the number of reads that would be filtered given a set of settings and prints this to the terminal.
Further, it tracks the number of singleton reads. The following metrics will always be tracked regardless of what you specify (the order output also matches this):

 * Total reads (including unmapped)
 * Mapped reads
 * Reads in blacklisted regions (--blackListFileName)

The following metrics are estimated according to the --binSize and --distanceBetweenBins parameters
 * Estimated mapped reads filtered (the total number of mapped reads filtered for any reason)
 * Alignments with a below threshold MAPQ (--minMappingQuality)
 * Alignments with at least one missing flag (--samFlagInclude)
 * Alignments with undesirable flags (--samFlagExclude)
 * Duplicates determined by sincei (--duplicateFilter)
 * Duplicates marked externally (e.g., by picard)
 * Singletons (paired-end reads with only one mate aligning)
 * Wrong strand (due to --filterRNAstrand)

The sum of these may be more than the total number of reads. Note that alignments are sampled from bins of size --binSize spaced --distanceBetweenBins apart.
""",
        usage="Example usage: scFilterStats.py -b sample1.bam sample2.bam -bc barcodes.txt > log.txt",
        add_help=False,
    )

    return parser


def getFiltered_worker(arglist):
    chrom, start, end, args = arglist
    # Fix the bounds
    if end - start > args.binSize and end - start > args.distanceBetweenBins:
        end -= args.distanceBetweenBins
    if end <= start:
        end = start + 1
    ## open genome if needed
    if args.genome2bit:
        twoBitGenome = py2bit.open(args.genome2bit, True)
    ## open blacklist file
    blackList = None
    if args.blackListFileName is not None:
        blackList = GTF(args.blackListFileName)

    o = []
    for fname in args.bamfiles:
        fh = bamHandler.openBam(fname)
        chromUse = utilities.mungeChromosome(chrom, fh.references)
        prev_pos = set()
        lpos = None

        ## a dict with barcodes = keys
        # metrics
        blacklisted = {}
        minMapq = {}
        samFlagInclude = {}
        samFlagExclude = {}
        internalDupes = {}
        externalDupes = {}
        singletons = {}
        filterRNAstrand = {}
        filterMotifs = {}
        filterGC = {}
        minAlignedFraction = {}
        # trackers
        nFiltered = {}
        total = {}  # This is only used to estimate the percentage affected
        filtered = {}

        for b in args.barcodes:
            total[b] = 0  # This is only used to estimate the percentage affected
            nFiltered[b] = 0
            blacklisted[b] = 0
            minMapq[b] = 0
            samFlagInclude[b] = 0
            samFlagExclude[b] = 0
            internalDupes[b] = 0
            externalDupes[b] = 0
            singletons[b] = 0
            filterRNAstrand[b] = 0
            filterMotifs[b] = 0
            filterGC[b] = 0
            minAlignedFraction[b] = 0

        for read in fh.fetch(chromUse, start, end):
            try:
                bc = read.get_tag(args.cellTag)
            except KeyError:
                continue
            # also keep a counter for barcodes not in whitelist?
            if bc not in args.barcodes:
                continue

            filtered[bc] = 0
            if read.pos < start:
                # ensure that we never double count (in case distanceBetweenBins == 0)
                continue

            if read.flag & 4:
                # Ignore unmapped reads, they were counted already
                continue

            if args.minMappingQuality and read.mapq < args.minMappingQuality:
                filtered[bc] = 1
                minMapq[bc] += 1
            if args.samFlagInclude and read.flag & args.samFlagInclude != args.samFlagInclude:
                filtered[bc] = 1
                samFlagInclude[bc] += 1
            if args.samFlagExclude and read.flag & args.samFlagExclude != 0:
                filtered[bc] = 1
                samFlagExclude[bc] += 1

            if args.minAlignedFraction:
                if not checkAlignedFraction(read, args.minAlignedFraction):
                    filtered[bc] = 1
                    minAlignedFraction[bc] += 1

            ## reads in blacklisted regions
            if blackList and blackList.findOverlaps(
                chrom,
                read.reference_start,
                read.reference_start + read.infer_query_length(always=False) - 1,
            ):
                filtered[bc] = 1
                blacklisted[bc] += 1

            ## Duplicates
            if args.duplicateFilter:
                tup = getDupFilterTuple(read, bc, args.duplicateFilter)
                if lpos is not None and lpos == read.reference_start and tup in prev_pos:
                    filtered[bc] = 1
                    internalDupes[bc] += 1
                if lpos != read.reference_start:
                    prev_pos.clear()
                lpos = read.reference_start
                prev_pos.add(tup)
            if read.is_duplicate:
                filtered[bc] = 1
                externalDupes[bc] += 1
            if read.is_paired and read.mate_is_unmapped:
                filtered[bc] = 1
                singletons[bc] += 1

            ## remove reads with low/high GC content
            if args.GCcontentFilter:
                if not checkGCcontent(read, args.GCcontentFilter[0], args.GCcontentFilter[1]):
                    filtered[bc] = 1
                    filterGC[bc] += 1

            ## remove reads that don't pass the motif filter
            if args.motifFilter:
                test = [checkMotifs(read, chrom, twoBitGenome, m[0], m[1]) for m in args.motifFilter]
                # if none given motif found, return true
                if not any(test):
                    filtered[bc] = 1
                    filterMotifs[bc] += 1

            # filterRNAstrand
            if args.filterRNAstrand:
                if read.is_paired:
                    if args.filterRNAstrand == "forward":
                        if read.flag & 144 == 128 or read.flag & 96 == 64:
                            pass
                        else:
                            filtered[bc] = 1
                            filterRNAstrand[bc] += 1
                    elif args.filterRNAstrand == "reverse":
                        if read.flag & 144 == 144 or read.flag & 96 == 96:
                            pass
                        else:
                            filtered[bc] = 1
                            filterRNAstrand[bc] += 1
                else:
                    if args.filterRNAstrand == "forward":
                        if read.flag & 16 == 16:
                            pass
                        else:
                            filtered[bc] = 1
                            filterRNAstrand[bc] += 1
                    elif args.filterRNAstrand == "reverse":
                        if read.flag & 16 == 0:
                            pass
                        else:
                            filtered[bc] = 1
                            filterRNAstrand[bc] += 1

            total[bc] += 1
            nFiltered[bc] += filtered[bc]
        fh.close()

        # first make a tuple where each entry is a dict of barcodes:value
        tup = (
            total,
            nFiltered,
            blacklisted,
            minMapq,
            samFlagInclude,
            samFlagExclude,
            internalDupes,
            externalDupes,
            singletons,
            filterRNAstrand,
            filterMotifs,
            filterGC,
            minAlignedFraction,
        )

        # now simplify it
        merged = {}
        for b in args.barcodes:
            merged[b] = tuple(merged[b] for merged in tup)
        # now merged is a dict with each key = barcode, values = tuple of stats
        # Now convert it to array
        out = np.stack([v for k, v in merged.items()])
        # out is an array with row = len(barcode) [384], column = len(stats) [11]
        o.append(out)
    return o


def main(args=None):
    args, rowLabels = ParserCommon.validateInputs(parseArguments().parse_args(args))
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    if args.outFile is None:
        of = sys.stdout
    else:
        of = open(args.outFile, "w")

    for bam in args.bamfiles:
        x = bamHandler.openBam(bam, returnStats=True, nThreads=args.numberOfProcessors)[0]
        chrom_sizes = list(zip(x.references, x.lengths))

        checkBAMtag(x, bam, args.cellTag)
        if args.groupTag:
            checkBAMtag(x, bam, args.groupTag)
            sys.stderr.write(
                "--groupTag is not implemented for scFilterStats yet! \
            Please split your BAM file by {} and re-run scFilterStats. \n".format(
                    args.groupTag
                )
            )
            exit(1)
        x.close()

    # Get the remaining metrics
    res = mapReduce(
        [args],
        getFiltered_worker,
        chrom_sizes,
        genomeChunkLength=args.binSize + args.distanceBetweenBins,
        blackListFileName=args.blackListFileName,
        numberOfProcessors=args.numberOfProcessors,
        verbose=args.verbose,
    )
    ## res, should be the list of np.arrays of length (len(barcodes) * 9)

    ## final output is an array where nrows = bamfiles*barcodes, ncol = No. of stats
    final_array = np.asarray(res).sum(axis=0)
    ## get final row/col Names (bamnames_barcode)
    colLabels = [
        "Total_sampled",
        "Filtered",
        "Blacklisted",
        "Low_MAPQ",
        "Missing_Flags",
        "Excluded_Flags",
        "Internal_Duplicates",
        "Marked_Duplicates",
        "Singletons",
        "Wrong_strand",
        "Wrong_motif",
        "Unwanted_GC_content",
        "Low_aligned_fraction",
    ]

    final_df = pd.DataFrame(data=np.concatenate(final_array), index=rowLabels, columns=colLabels)
    final_df.index.name = "Cell_ID"
    ## since stats are approximate, present results as %
    final_df.iloc[:, 1:] = final_df.iloc[:, 1:].div(final_df.Total_sampled, axis=0) * 100

    if args.outFile is not None:
        final_df.to_csv(args.outFile, sep="\t")
    else:
        print(final_df)

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys
import itertools
import numpy as np
import pandas as pd

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei import ParserCommon
from sincei import WriteBedGraph

debug = 0


def parseArguments():
    io_args = ParserCommon.inputOutputOptions(opts=["bamfiles", "groupInfo", "outFilePrefix"])
    bam_args = ParserCommon.bamOptions(default_opts={"binSize": 100})
    read_args = ParserCommon.readOptions()
    filter_args = ParserCommon.filterOptions()
    other_args = ParserCommon.otherOptions()
    parser = argparse.ArgumentParser(
        parents=[io_args, get_args(), bam_args, filter_args, read_args, other_args],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="This tool takes alignments of reads or fragments "
        "as input (BAM files), along with cell grouping information, such as "
        "barcode -> batch, or barcode -> cluster, as tsv file, and generates a "
        "coverage track (bigWig or bedGraph) per group as output. "
        "The coverage is calculated as the number of reads per bin, "
        "where bins are short consecutive counting windows of a defined "
        "size. It is possible to extended/change the length of the reads "
        "to better reflect the actual fragment length. *scBulkCoverage* "
        "offers normalization per cluster using different methods \n",
        usage="An example usage is:"
        "$ scBulkCoverage -b file1.bam file2.bam --labels file1 file2 -g scClusterCells_output.tsv -o coverage.bw",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)

    optional = parser.add_argument_group("Coverage-related Options")
    optional.add_argument(
        "--outFileFormat",
        "-of",
        help='Output file type. Either "bigwig" or "bedgraph".',
        choices=["bigwig", "bedgraph"],
        default="bigwig",
    )

    optional.add_argument(
        "--normalizeUsing",
        "-n",
        help="How to normalize the pseudo-bulk counts. Options are "
        '"CPM": normalized each bin to the counts per million mapped reads in that group.\n'
        '"Frequency": binarize the coverage per bin and normalize to the total no. of cells per group. \n'
        '"Mean": get mean signal per bin across cells in each group.\n'
        '"None": simply return the sum of coverage per group.',
        choices=["CPM", "Frequency", "Mean", "None"],
        default="CPM",
    )

    optional.add_argument(
        "--ignoreForNormalization",
        "-ig",
        help="Chromosomes to skip while calculating normalization factors",
        nargs="+",
        default=None,
    )

    optional.add_argument(
        "--normalizeByReference",
        "-nr",
        help="NOT IMPLEMENTED: Normalize each group of cells by a reference group (which must be present in the --groupinfo file)"
        "Note that the --normalizeUsing method is applied beforehand.",
        choices=["ratio", "log2_ratio", "difference", "None"],
        default=None,
    )

    optional.add_argument(
        "--scaleFactor",
        help="The computed scaling factor (or 1, if not applicable) will "
        "be multiplied by this. (Default: %(default)s)",
        default=1.0,
        type=float,
        required=False,
    )

    optional.add_argument(
        "--MNase",
        help="Determine nucleosome positions from MNase-seq/CUTnRUN data. "
        "Only 3 nucleotides at the center of each fragment are counted. "
        "The fragment ends are defined by the two mate reads. Only fragment lengths"
        "between 130 - 200 bp are considered to avoid dinucleosomes or other artifacts. "
        "By default, any fragments smaller or larger than this are ignored. To "
        "over-ride this, use the --minFragmentLength and --maxFragmentLength options, "
        "which will default to 130 and 200 if not otherwise specified in the presence "
        "of --MNase. *NOTE*: Requires paired-end data. A bin size of 1 is recommended.",
        action="store_true",
    )

    optional.add_argument(
        "--Offset",
        help="Uses this offset inside of each read as the signal. This is useful in "
        "cases like RiboSeq or GROseq, where the signal is 12, 15 or 0 bases past the "
        "start of the read. This can be paired with the --filterRNAstrand option. "
        "Note that negative values indicate offsets from the end of each read. A value "
        "of 1 indicates the first base of the alignment (taking alignment orientation "
        "into account). Likewise, a value of -1 is the last base of the alignment. An "
        "offset of 0 is not permitted. If two values are specified, then they will be "
        "used to specify a range of positions. Note that specifying something like "
        "--Offset 5 -1 will result in the 5th through last position being used, which "
        "is equivalent to trimming 4 bases from the 5-prime end of alignments. Note "
        "that if you specify --centerReads, the centering will be performed before the "
        "offset.",
        metavar="INT",
        type=int,
        nargs="+",
        required=False,
    )

    return parser


def main(args=None):
    args = ParserCommon.validateInputs(
        parseArguments().parse_args(args), process_barcodes=False
    )  # newlabels are discarded, since they are re-created from groupInfo file
    global debug
    if args.verbose:
        sys.stderr.write("Specified --scaleFactor: {}\n".format(args.scaleFactor))
        debug = 1
    else:
        debug = 0
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    ## read the group info file (TODO: write a validator)
    df_err = """ *Error*:No. of columns in --groupInfo file not recognized.
             Please provide either 3 (sample, barcode, group) or 4 (sample::barcode, umap1, umap2, group) column file"""
    ## if the no. of columns are 3, expect "sample", "barcode", "cluster", if 4, expect sample:bc, umap1, umap2, cluster
    df = pd.read_csv(args.groupInfo, sep="\t", index_col=None, comment="#")
    if len(df.columns) == 3:
        df.columns = ["sample", "barcode", "cluster"]
    elif len(df.columns) == 4:
        df.columns = ["Cell_ID", "umap1", "umap2", "cluster"]
        try:
            df[["sample", "barcode"]] = df.Cell_ID.str.split("::", expand=True)
        except:
            sys.exit(df_err)
    else:
        sys.exit(df_err)

    df.index = df[["sample", "barcode"]].apply(lambda x: "::".join(x), axis=1)
    # barcodes = groupInfo.barcode.unique().tolist()

    ## match the sample labels with groupInfo labels before proceeding
    ## create new DF with user-provided bam+labels (this would match the counts)

    # check that sample names in df and --labels match
    labels_not_in_df = set(args.labels).difference(set(df["sample"]))
    df_not_in_labels = set(df["sample"]).difference(set(args.labels))
    if bool(df_not_in_labels):
        sys.stderr.write(
            "Some (or all) of the samples indicated in the groupInfo file "
            "are absent from the bam file labels! \n"
            "Mismatched samples are: {} \n".format(df_not_in_labels)
        )
        df = df.loc[df["sample"].isin(set(args.labels)),]
    elif bool(labels_not_in_df):
        sys.stderr.write(
            "Some (or all) of the samples indicated in --labels "
            "are absent from the in the groupInfo file! \n"
            "Mismatched samples are: {} \n".format(labels_not_in_df)
        )

    # make another df, containing union of all barcodes, and in the same order per sample
    barcodes = df["barcode"].unique().tolist()
    sm = list(itertools.chain.from_iterable(itertools.repeat(lab, len(barcodes)) for lab in args.labels))
    bc = barcodes * len(args.labels)
    groupInfo = pd.DataFrame({"sample": sm, "barcode": bc})
    groupInfo.index = groupInfo[["sample", "barcode"]].apply(lambda x: "::".join(x), axis=1)
    groupInfo = pd.merge(
        groupInfo,
        df["cluster"],
        how="left",
        left_index=True,
        right_index=True,
        sort=False,
    )
    groupInfo = groupInfo.reset_index()[["sample", "barcode", "cluster"]]
    groupInfo["cluster"] = groupInfo["cluster"].astype("category")
    # re-construct new labels (sample+bc)
    newlabels = ["::".join([x, y]) for x, y in zip(groupInfo["sample"], groupInfo["barcode"])]
    # Normalization options
    scale_factor = args.scaleFactor
    func_args = {"scaleFactor": args.scaleFactor}

    coverageAsFrequency = False
    if not args.ignoreForNormalization:
        args.ignoreForNormalization = []
    if args.normalizeUsing == "None":
        args.normalizeUsing = None  # For the sake of sanity
    elif args.normalizeUsing == "Frequency":
        coverageAsFrequency = True
        args.normalizeUsing = "Mean"  # mean of binarized counts shall give us frequency

    # This fixes issue #520, where --extendReads wasn't honored if --filterRNAstrand was used
    if args.filterRNAstrand and not args.Offset:
        args.Offset = [1, -1]

    if args.MNase:
        # check that library is paired end
        # using getFragmentAndReadSize
        from deeptools.getFragmentAndReadSize import get_read_and_fragment_length

        fraglengths = [
            get_read_and_fragment_length(
                file,
                return_lengths=False,
                blackListFileName=args.blackListFileName,
                numberOfProcessors=args.numberOfProcessors,
                verbose=args.verbose,
            )
            for file in args.bamfiles
        ]
        if any([x[0] is None for x in fraglengths]):
            sys.exit("*Error*: For the --MNAse function a paired end library is required. ")

        # Set some default fragment length bounds
        if args.minFragmentLength == 0:
            args.minFragmentLength = 130
        if args.maxFragmentLength == 0:
            args.maxFragmentLength = 200

        wr = CenterFragment(
            args.bamfiles,
            binLength=args.binSize,
            stepSize=args.binSize,
            barcodes=barcodes,
            clusterInfo=groupInfo,
            cellTag=args.cellTag,
            groupTag=args.groupTag,
            groupLabels=newlabels,
            motifFilter=args.motifFilter,
            genome2bit=args.genome2bit,
            GCcontentFilter=args.GCcontentFilter,
            region=args.region,
            blackListFileName=args.blackListFileName,
            numberOfProcessors=args.numberOfProcessors,
            extendReads=args.extendReads,
            minMappingQuality=args.minMappingQuality,
            duplicateFilter=args.duplicateFilter,
            center_read=args.centerReads,
            zerosToNans=False,
            samFlag_include=args.samFlagInclude,
            samFlag_exclude=args.samFlagExclude,
            minFragmentLength=args.minFragmentLength,
            maxFragmentLength=args.maxFragmentLength,
            chrsToSkip=args.ignoreForNormalization,
            binarizeCoverage=coverageAsFrequency,
            verbose=args.verbose,
        )

    elif args.Offset:
        if len(args.Offset) > 1:
            if args.Offset[0] == 0:
                sys.exit(
                    "*Error*: An offset of 0 isn't allowed, since offsets are 1-based positions inside each alignment."
                )
            if args.Offset[1] > 0 and args.Offset[1] < args.Offset[0]:
                sys.exit("'Error*: The right side bound is less than the left-side bound. This is inappropriate.")
        else:
            if args.Offset[0] == 0:
                sys.exit(
                    "*Error*: An offset of 0 isn't allowed, since offsets are 1-based positions inside each alignment."
                )
        wr = OffsetFragment(
            args.bamfiles,
            binLength=args.binSize,
            stepSize=args.binSize,
            barcodes=barcodes,
            clusterInfo=groupInfo,
            cellTag=args.cellTag,
            groupTag=args.groupTag,
            groupLabels=newlabels,
            motifFilter=args.motifFilter,
            genome2bit=args.genome2bit,
            GCcontentFilter=args.GCcontentFilter,
            region=args.region,
            numberOfProcessors=args.numberOfProcessors,
            extendReads=args.extendReads,
            minMappingQuality=args.minMappingQuality,
            duplicateFilter=args.duplicateFilter,
            center_read=args.centerReads,
            zerosToNans=False,
            samFlag_include=args.samFlagInclude,
            samFlag_exclude=args.samFlagExclude,
            minFragmentLength=args.minFragmentLength,
            maxFragmentLength=args.maxFragmentLength,
            chrsToSkip=args.ignoreForNormalization,
            binarizeCoverage=coverageAsFrequency,
            verbose=args.verbose,
        )
        wr.filter_strand = args.filterRNAstrand
        wr.Offset = args.Offset
    else:
        wr = WriteBedGraph.WriteBedGraph(
            args.bamfiles,
            binLength=args.binSize,
            stepSize=args.binSize,
            barcodes=barcodes,
            clusterInfo=groupInfo,
            cellTag=args.cellTag,
            groupTag=args.groupTag,
            groupLabels=newlabels,
            motifFilter=args.motifFilter,
            genome2bit=args.genome2bit,
            GCcontentFilter=args.GCcontentFilter,
            region=args.region,
            blackListFileName=args.blackListFileName,
            numberOfProcessors=args.numberOfProcessors,
            extendReads=args.extendReads,
            minMappingQuality=args.minMappingQuality,
            duplicateFilter=args.duplicateFilter,
            center_read=args.centerReads,
            zerosToNans=False,
            samFlag_include=args.samFlagInclude,
            samFlag_exclude=args.samFlagExclude,
            minFragmentLength=args.minFragmentLength,
            maxFragmentLength=args.maxFragmentLength,
            chrsToSkip=args.ignoreForNormalization,
            binarizeCoverage=coverageAsFrequency,
            verbose=args.verbose,
        )

    wr.run(
        WriteBedGraph.scaleCoverage,
        func_args,
        args.outFilePrefix,
        blackListFileName=args.blackListFileName,
        normUsing=args.normalizeUsing,
        format=args.outFileFormat,
        smoothLength=None,
    )


class OffsetFragment(WriteBedGraph.WriteBedGraph):
    """
    Class to redefine the get_fragment_from_read for the --Offset case
    """

    def filterStrand(self, read, rv):
        """
        A generic read filtering function that gets used by everything in this class.

        rv is returned if the strand is correct, otherwise [(None, None)]
        """
        # Filter by RNA strand, if desired
        if read.is_paired:
            if self.filter_strand == "forward":
                if read.flag & 144 == 128 or read.flag & 96 == 64:
                    return rv
            elif self.filter_strand == "reverse":
                if read.flag & 144 == 144 or read.flag & 96 == 96:
                    return rv
            else:
                return rv
        else:
            if self.filter_strand == "forward":
                if read.flag & 16 == 16:
                    return rv
            elif self.filter_strand == "reverse":
                if read.flag & 16 == 0:
                    return rv
            else:
                return rv

        return [(None, None)]

    def get_fragment_from_read_list(self, read, offset):
        """
        Return the range of exons from the 0th through 1st bases, inclusive. Positions are 1-based
        """
        rv = [(None, None)]
        blocks = read.get_blocks()
        blockLen = sum([x[1] - x[0] for x in blocks])

        if self.defaultFragmentLength != "read length":
            if self.is_proper_pair(read, self.maxPairedFragmentLength):
                if read.is_reverse:
                    foo = (read.next_reference_start, read.reference_start)
                    if foo[0] < foo[1]:
                        blocks.insert(0, foo)
                else:
                    foo = (
                        read.reference_end,
                        read.reference_end + abs(read.template_length) - read.infer_query_length(),
                    )
                    if foo[0] < foo[1]:
                        blocks.append(foo)

            # Extend using the default fragment length
            else:
                if read.is_reverse:
                    foo = (
                        read.reference_start - self.defaultFragmentLength + read.infer_query_length(),
                        read.reference_start,
                    )
                    if foo[0] < 0:
                        foo = (0, foo[1])
                    if foo[0] < foo[1]:
                        blocks.insert(0, foo)
                else:
                    foo = (
                        read.reference_end,
                        read.reference_end + self.defaultFragmentLength - read.infer_query_length(),
                    )
                    if foo[0] < foo[1]:
                        blocks.append(foo)

        stretch = []
        # For the sake of simplicity, convert [(10, 20), (30, 40)] to [10, 11, 12, 13, ..., 40]
        # Then subset accordingly
        for block in blocks:
            stretch.extend(range(block[0], block[1]))
        if read.is_reverse:
            stretch = stretch[::-1]

        # Handle --centerReads
        if self.center_read:
            _ = (len(stretch) - blockLen) // 2
            stretch = stretch[_ : _ + blockLen]

        # Subset by --Offset
        try:
            foo = stretch[offset[0] : offset[1]]
        except:
            return rv

        if len(foo) == 0:
            return rv
        if read.is_reverse:
            foo = foo[::-1]

        # Convert the stretch back to a list of tuples
        foo = np.array(foo)
        d = foo[1:] - foo[:-1]
        idx = np.argwhere(d > 1).flatten().tolist()  # This now holds the interval bounds as a list
        idx.append(-1)
        last = 0
        rv = []
        for i in idx:
            rv.append((foo[last].astype("int"), foo[i].astype("int") + 1))
            last = i + 1

        # Handle strand filtering, if needed
        return self.filterStrand(read, rv)

    def get_fragment_from_read(self, read):
        """
        This is mostly a wrapper for self.get_fragment_from_read_list(),
        which needs a list and for the offsets to be tweaked by 1.
        """
        offset = [x for x in self.Offset]
        if len(offset) > 1:
            if offset[0] > 0:
                offset[0] -= 1
            if offset[1] < 0:
                offset[1] += 1
        else:
            if offset[0] > 0:
                offset[0] -= 1
                offset = [offset[0], offset[0] + 1]
            else:
                if offset[0] < -1:
                    offset = [offset[0], offset[0] + 1]
                else:
                    offset = [offset[0], None]
        if offset[1] == 0:
            # -1 gets switched to 0, which screws things up
            offset = (offset[0], None)
        return self.get_fragment_from_read_list(read, offset)


class CenterFragment(WriteBedGraph.WriteBedGraph):
    """
    Class to redefine the get_fragment_from_read for the --MNase case

    The coverage of the fragment is defined as the 2 or 3 basepairs at the
    center of the fragment length.
    """

    def get_fragment_from_read(self, read):
        """
        Takes a proper pair fragment of high quality and limited
        to a certain length and outputs the center
        """
        fragment_start = fragment_end = None

        # only paired forward reads are considered
        # Fragments have already been filtered according to length
        if read.is_proper_pair and not read.is_reverse and 1 < abs(read.tlen):
            # distance between pairs is even return two bases at the center
            if read.tlen % 2 == 0:
                fragment_start = read.pos + read.tlen / 2 - 1
                fragment_end = fragment_start + 2

            # distance is odd return three bases at the center
            else:
                fragment_start = read.pos + read.tlen / 2 - 1
                fragment_end = fragment_start + 3

        return [(fragment_start, fragment_end)]



# topic models
import numpy as np
import pandas as pd
from gensim import corpora, matutils, models
import copy
from sklearn.preprocessing import binarize

### ------ Functions ------


class TOPICMODEL:
    r"""
    Computes LSA or LDA for a given matrix and returns the cell-topic matrix.

    Parameters
    ----------
    adata : AnnData
        AnnData object containing the data matrix in adata.X, with cells in adata.obs_names and regions in adata.var_names.
    n_topics : int
        Number of Topics / Principal Components for modeling.
    binarize : bool, optional
        If True, the input matrix will be binarized (default is False). Recommended for LDA.
    smart_code : str
        SMART (System for the Mechanical Analysis and Retrieval of Text) code for weighting of input matrix for TFIDF.
        Only valid for the LSA model. The default ("lfu") corresponds to "log"TF * IDF, and "pivoted unique" normalization of document length. For more information, see: https://en.wikipedia.org/wiki/SMART_Information_Retrieval_System
    n_passes : int, optional
        Number of passes for the LDA model. Default is 1.
    n_workers : int, optional
        Number of workers for the LDA model. Default is 1.

    Returns
    -------
    An object of class TOPICMODEL.
    """

    def __init__(
        self,
        adata,
        n_topics,
        binarize=False,
        smart_code="lfu",
        n_passes=1,
        n_workers=1,
    ):
        self.cells = adata.obs_names.to_list()
        self.regions_dict = corpora.dictionary.Dictionary([adata.var_names.to_list()])
        mtx = adata.X.copy().transpose()
        if binarize:
            mtx = binarize(mtx, copy=True)
        self.corpus = matutils.Sparse2Corpus(mtx)
        self.shape = adata.shape
        self.n_topics = n_topics
        self.smart_code = smart_code
        self.n_passes = n_passes
        self.n_workers = n_workers
        self.lsi_model = None
        self.lda_model = None
        self.cell_topic_dist = None
        self.topic_region_dist = None

    def runLSA(self):
        r"""
        Computes LSA for a given matrix and returns the updated object
        """

        # LSA
        tfidf = models.TfidfModel(self.corpus, id2word=self.regions_dict, normalize=True, smartirs=self.smart_code)
        self.corpus_tfidf = tfidf[self.corpus]
        self.lsi_model = models.LsiModel(self.corpus_tfidf, id2word=self.regions_dict, num_topics=self.n_topics)
        self.cell_topic_dist = self.lsi_model[
            self.corpus_tfidf
        ]  # lsi[X] computes U^-1*X, which equals V*S (its shape is num_docs * num_topics).

        # Compute Coherence Score
        coherence_model_lsa = models.CoherenceModel(
            model=self.lsi_model, corpus=self.corpus, dictionary=self.regions_dict, coherence="u_mass"
        )
        coherence_lsa = coherence_model_lsa.get_coherence()
        print("\nCoherence Score: ", coherence_lsa)

    def runLDA(self):
        r"""
        Computes LDA model for a given matrix and returns the updated object
        """

        self.lda_model = models.LdaMulticore(
            corpus=self.corpus, num_topics=self.n_topics, passes=self.n_passes, workers=self.n_workers
        )
        # get topic distributions for each document
        self.cell_topic_dist = self.lda_model[self.corpus]
        # get topic-word distributions
        self.topic_region_dist = self.lda_model.get_topics()

    def get_cell_topic(self, pop_sparse_cells=False):
        r"""
        Get cell-topic matrix from the updated object

        Returns
        -------
        cell_topic : pandas dataframe
            Cell-topic matrix
        """

        cells = copy.deepcopy(self.cells)
        ## make cell-topic df
        li = [[tup[0] for tup in x] for x in self.cell_topic_dist]
        li_val = [[tup[1] for tup in x] for x in self.cell_topic_dist]

        # if all documents don't have same set of topics, (optionally) remove them
        if len(set([len(x) for x in li_val])) > 1:
            bad_idx = sorted([i for i, v in enumerate(li_val) if len(v) != self.n_topics], reverse=True)
            print(f"{len(bad_idx)} cells were detected which don't contribute to all {self.n_topics} topics.")
            if pop_sparse_cells:
                print("Removing these cells from the analysis")
                for x in bad_idx:
                    li_val.pop(x)
                    li.pop(x)
                    cells.pop(x)

                li_val = np.stack(li_val)
                cell_topic = pd.DataFrame(li_val, columns=[f"topic_{x}" for x in range(self.n_topics)])
            else:
                cell_topic = np.zeros((len(li_val), self.n_topics))
                for i, v in enumerate(li_val):
                    for j, val in enumerate(v):
                        print(f"Index [{i}, {li[i][j]}] = {val}")
                        cell_topic[i, li[i][j]] = val
                cell_topic = pd.DataFrame(cell_topic, columns=[f"topic_{x}" for x in range(self.n_topics)])

        cell_topic.index = cells

        return cell_topic



import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt

import scipy as sp
import numpy as np
import pandas as pd


## Fast Fourier Transform
def ffttable(selected):
    r"""Computes the FFT of the fragment length distribution

    Parameters
    ----------
    selected : numpy array
        Array of fragment lengths

    Returns
    -------
    pandas dataframe
        Dataframe with two columns. The first column contains the frequencies
        and the second column the power spectrum.

    Examples
    --------
    >>> test = Tester()
    >>> c = CountReadsPerBin([test.bamFile1, test.bamFile2], 50, 4)
    >>> num_reads_per_bin, regionList = c.run()
    >>> selected = num_reads_per_bin[:,0]
    >>> ffttable(selected)
       freq         value
    0   0.0  7.812500e+06
    1   1.0  1.812500e+06
    2   2.0  1.812500e+06
    3   3.0  1.812500e+06
    4   4.0  1.812500e+06
    5   5.0  1.812500
    """
    selected = np.log2(selected + 1)
    selected2 = [y - x for x, y in zip(selected, selected[1:])]
    fragment_fft = sp.fftpack.fft(selected2)
    fragment_psd = np.abs(fragment_fft) ** 2
    fftfreq = sp.fftpack.fftfreq(len(fragment_psd), 10.0)
    d = pd.DataFrame({"freq": fftfreq, "value": fragment_psd})
    d2 = d[d.freq > 0]
    return d2


# get fragment size disribution and fft value from dict{barcode:fragment_size_list}
def fragment_distribution(fragment_len_dict, length_plot):
    r"""Plot fragment length distribution

    Parameters
    ----------
    fragment_len_dict : dict
        Dictionary containing the fragment length for each barcode.
    length_plot : str
        File name to save the plot.

    Returns
    -------
    dict
        Dictionary containing the top 20 frequencies for each barcode.


    Examples
    --------
    >>> test = Tester()
    >>> fragment_len_dict = {'AAACCTGAGAGGTTCT': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
    ...                      'AAACCTGAGAGGTTCT': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000]}
    >>> fragment_distribution(fragment_len_dict, test.length_plot)
    """
    plt.style.use("classic")
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)

    outdict = dict.fromkeys(fragment_len_dict.keys())
    for barcode in outdict.keys():
        fragment_len = fragment_len_dict[barcode]

        n, bins, patches = ax1.hist(fragment_len, bins=100, color="orange", range=(0, 1000), alpha=0.3)
        # calculating fft
        dflist = []
        for num in range(80, 101, 1):
            dflist.append(ffttable(n[10:num]))
        d2 = pd.concat(dflist, ignore_index=True)
        d2 = d2.sort_values(by=["freq"], ascending=False)
        outdict[barcode] = d2

    # plot fragment sizes
    ax1.xaxis.set_ticks_position("bottom")
    ax1.yaxis.set_ticks_position("left")
    plt.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    plt.xlabel("Fragment length", fontsize=20)
    plt.ylabel("Fragment count", fontsize=20)
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    ax1.set_title("Fragment distribution", fontsize=20)
    plt.savefig(length_plot, dpi=300, bbox_inches="tight")
    plt.close(fig)

    return outdict


## plot the fragment periodicity per barcode using the output of
def fftplot(outdict, plot):
    r"""Computes the periodicity of the fragment distribution

    Parameters
    ----------
    outdict : dict
        Dictionary containing the fragment distribution for each barcode
    plot : str
        File name to save the plot

    Returns
    -------
    dict
        Dictionary containing the periodicity of the fragment distribution for each barcode

    Examples
    --------
    >>> test = Tester()
    >>> d = test.get_fragment_distribution()
    >>> fftplot(d, 'test.png')
    """

    plt.style.use("classic")
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    periodicity = dict.fromkeys(outdict.keys())
    for barcode in outdict.keys():
        d2 = outdict[barcode]
        max_y_pos = 1 / d2.loc[d2.value.idxmax(), "freq"]
        mean_value = np.mean(d2.value)

        ## color barcodes where max periodicity is not between 120 and 160, or 240 and 320
        p1 = range(120, 160, 1)
        p2 = range(240, 320, 1)

        mononuc_periodicity = np.mean(d2.loc[[int(x) in p1 for x in 1 / d2.freq], "value"])
        dinuc_periodicity = np.mean(d2.loc[[int(x) in p2 for x in 1 / d2.freq], "value"])

        col = "grey"
        ax.plot(1 / d2.freq, 10 * np.log10(d2.value + 1), color=col, alpha=0.2)

        periodicity[barcode] = [mononuc_periodicity, dinuc_periodicity]

    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    ax.set_xlim(0, 400)
    ax.set_xlabel("Period (bp)", fontsize=20)
    ax.set_ylabel("Power", fontsize=20)
    ax.set_title("Period of fragment distribution", fontsize=20)

    plt.savefig(plot, dpi=300, bbox_inches="tight")
    return periodicity



import numpy as np
import torch
import torch.optim
from torch.utils.data import TensorDataset, DataLoader
from tqdm import tqdm
import anndata as ad

try:
    import mctorch.nn as mnn
    import mctorch.optim as moptim
except ImportError:
    raise ImportError("Please install mctorch package via `pip install --user mctorch")

from sincei.ExponentialFamily import Gaussian, Poisson, Bernoulli, Beta, Gamma, LogNormal, SigmoidBeta

EXPONENTIAL_FAMILY_DICT = {
    "gaussian": Gaussian,
    "poisson": Poisson,
    "bernoulli": Bernoulli,
    "beta": Beta,
    "gamma": Gamma,
    "lognormal": LogNormal,
    "log_normal": LogNormal,
    "sigmoid_beta": SigmoidBeta,
}
LEARNING_RATE_LIMIT = 10 ** (-10)


class GLMPCA:
    r"""Performs GLM-PCA on a data matrix to reduce dimensionality

    This class computes the generalized-linear model principal components (GLM-PC)
    of a dataset by exploiting the framework of saturated parameters.
    Specifically, given an exponential family distribution choosen following some
    prior knowledge, GLM-PCA will find a collection of directions which maximise the
    reconstruction error, computed as the negative log-likelihood of the chosen
    exponential family.

    By making use of an alternative formulation, our implementation can exploit
    automatic differentiation and can therefore rely on mini-batch Stochastic
    Gradient Descent. As a consequence, it scales to very large dataset.

    Another interesting feature of our implementation is that it does not require
    cumbersome Lagrangian derivations. If you wish to test an exponential family
    distribution not present in our implementation, adding a class in ExponentialFamily
    with the different density functionals defined there would suffice to use GLM-PCA.

    Parameters
    ----------
    n_pc : int
        Number of principal components.

    family: str
        Name of the exponential family distribution. Possible families: "gaussian", "poisson",
        "bernoulli", "beta", "gamma", "log_normal", "log_beta":, "sigmoid_beta". Default to
        "gaussian"

    family_params : dict
        Dictionary with family parameters to be added. List of parameters depend on the
        specific ExponentialFamily class chosen. Examples:
            - "n_jobs" (int) for parallelization, specifically for "beta" and "gamma".
            - "min_val" (float) for truncating in "poisson" or "beta".
            - "eps" (float) for convergence in inverse computation in "beta".
        Default to None.

    max_iter : int
        Maximum number of epochs in the GLM-PCA optimisation. Default to 100.

    learning_rate: float
        Learning rate to be used in the GLM-PCA optimisation. If learning_rate is too
        high and lead to NaN, our implementation automatically restarts the optimisation
        with a smaller value. Default to 0.2.


    batch_size : int
        Size of the batch in the SGD optimisation step. Default to 256.

    step_size: int
        Step size in optimiser scheduler. See more: https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.StepLR.html
        Default to 20.

    gamma: int
        Reduction parameter for optimiser scheduler. See more: https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.StepLR.html
        Default to 0.5

    n_init: int
        Number of GLM-PCA initializations. Useful if you want to explore different
        random seeds and starting points. Default to 1.

    init: str
        Method to initialize loadings. "spectral" performs SVD on  the saturated parameters from
        a small random batch of the dataset, "random" performs a random initialization on the
        Stiefel manifold. Default to "spectral".

    n_jobs: int
        Number of jobs used in parallel operations. Default to 1.
    """

    def __init__(
        self,
        n_pc,
        family="gaussian",
        family_params=None,
        max_iter=100,
        learning_rate=0.2,
        batch_size=256,
        step_size=20,
        gamma=0.5,
        n_init=1,
        init="spectral",
        n_jobs=1,
    ):
        self.n_pc = n_pc
        self.family = family
        self.family_params = family_params
        self.log_part_theta_matrices_ = None
        self.max_iter = np.abs(max_iter)
        self.learning_rate_ = learning_rate
        self.initial_learning_rate_ = learning_rate
        self.n_jobs = n_jobs
        self.batch_size = batch_size
        self.n_init = n_init
        self.gamma = gamma
        self.step_size = step_size
        self.init = init

        self.saturated_loadings_ = None
        # saturated_intercept_: before projecting
        self.saturated_intercept_ = None
        # reconstruction_intercept: after projecting
        self.reconstruction_intercept_ = None

        # Whether to perform sample or gene projection
        self.sample_projection = False

        self.exp_family_params = None
        self.loadings_learning_scores_ = []
        self.loadings_learning_rates_ = []

        # Initialize device
        self.device = None

        # Set up exponential family
        if type(family) is str:
            self.exponential_family = EXPONENTIAL_FAMILY_DICT[family](self.family_params)
        else:
            self.exponential_family = family

    def fit(self, X):
        r"""Fits a GLM-PCA to a specific dataset.

        Parameters
        ----------
        X : torch.Tensor, np.array or anndata object
            Dataset with cells in the rows and features in the columns.

        Returns
        -------
        bool
            Returns True if the fitting procedure has been successful.
        """
        if isinstance(X, ad.AnnData):
            X_fit = torch.Tensor(X.X.transpose().toarray() if hasattr(X.X, "toarray") else X.X.transpose())
        elif isinstance(X, np.ndarray):
            X_fit = torch.Tensor(X)
        elif isinstance(X, torch.Tensor):
            X_fit = X.clone()
        else:
            raise ValueError("X format unrecognised: %s != np.ndarray or torch.Tensor" % (type(X)))

        # Fit exponential family params (e.g., dispersion for negative binomial)
        self.exponential_family.initialize_family_parameters(X_fit)

        # Compute saturated parameters, alongside exponential family parameters (if needed)
        saturated_parameters = self.exponential_family.invert_g(X_fit)

        # Use saturated parameters to find loadings by projected gradient descent
        self.saturated_loadings_ = []
        self.saturated_intercept_ = []

        # Initialize the learning procedure
        self.learning_rate_ = self.initial_learning_rate_
        self.loadings_learning_scores_ = []
        self.loadings_learning_rates_ = []

        # Compute loadings for different parameters
        for _ in range(self.n_init):
            self._compute_saturated_loadings(X_fit, saturated_parameters)

        # Select best model
        training_cost = torch.Tensor(
            [
                self._optim_cost(loadings, intercept, X_fit, saturated_parameters)
                for loadings, intercept in zip(self.saturated_loadings_, self.saturated_intercept_)
            ]
        )
        best_model_idx = torch.argmin(training_cost)
        self.saturated_intercept_ = self.saturated_intercept_[best_model_idx]
        self.saturated_loadings_ = self.saturated_loadings_[best_model_idx]

        return True

    def transform(self, X):
        r"""Transforms and project dataset X onto the principal components.

        Parameters
        ----------
        X : torch.Tensor or np.array
            Dataset with cells in the rows and features in the columns.

        Returns
        -------
        torch.Tensor
            Projected saturated parameters.
        """
        saturated_parameters = self.exponential_family.invert_g(X)

        # Compute intercept term
        n = X.shape[0]
        intercept_term = self.saturated_intercept_.unsqueeze(0).repeat(n, 1).to(self.device)

        projected_parameters = saturated_parameters - intercept_term
        projected_parameters = projected_parameters.matmul(self.saturated_loadings_)

        return projected_parameters

    def _compute_saturated_loadings(self, X, saturated_parameters):
        # Runs one optimisation of the loadings, using mcTorch.
        loadings_ = self._saturated_loading_iter(saturated_parameters, X)
        self.saturated_loadings_.append(loadings_[0])
        self.saturated_intercept_.append(loadings_[1])

    def _saturated_loading_iter(self, saturated_parameters: torch.Tensor, X: torch.Tensor):
        r"""Computes the loadings solution of the GLM-PCA optimisation problem.

        Parameters
        ----------
        saturated_parameters : torch.Tensor
            Saturated parameters of the dataset X ($g^{-1}\left(X\right)$)
        X : torch.Tensor
            Dataset with cells in the rows and features in the columns.

        Returns
        -------
        torch.Tensor
            Projected saturated parameters.
        """
        if self.learning_rate_ < LEARNING_RATE_LIMIT:
            raise ValueError("LEARNING RATE IS TOO SMALL : DID NOT CONVERGE")

        # Set device for GPU usage
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Set up list of saving scores
        self.loadings_learning_scores_.append([])
        self.loadings_learning_rates_.append([])

        _optimizer, _loadings, _intercept, _lr_scheduler = self._create_saturated_loading_optim(
            parameters=saturated_parameters.data.clone(), X=X
        )

        # Load dataset
        train_data = TensorDataset(X, saturated_parameters.data.clone())
        train_loader = DataLoader(dataset=train_data, batch_size=self.batch_size, shuffle=True, drop_last=True)

        # Run epoch in a for loop
        self._loadings_epochs = [_loadings.clone().detach()]
        self._intercept_epochs = [_intercept.clone().detach()]
        for _ in tqdm(range(self.max_iter)):
            loss_val = []
            for batch_data, batch_parameters in train_loader:
                cost_step = self._optim_cost(
                    loadings=_loadings,
                    intercept=_intercept,
                    batch_data=batch_data,
                    batch_parameters=batch_parameters,
                )

                if "cuda" in str(self.device):
                    self.loadings_learning_scores_[-1].append(cost_step.cpu().detach().numpy())
                else:
                    self.loadings_learning_scores_[-1].append(cost_step.detach().numpy())
                cost_step.backward()
                _optimizer.step()
                _optimizer.zero_grad()
                self.loadings_learning_rates_[-1].append(_lr_scheduler.get_last_lr())
            _lr_scheduler.step()

            self._loadings_epochs.append(_loadings.clone().detach())
            self._intercept_epochs.append(_intercept.clone().detach())

            # If NaN or Inf is found in the parameters, start over optimisation with reduced learning rate.
            if np.isinf(self.loadings_learning_scores_[-1][-1]) or np.isnan(self.loadings_learning_scores_[-1][-1]):
                print("\tRESTART BECAUSE INF/NAN FOUND", flush=True)
                self.learning_rate_ = self.learning_rate_ * self.gamma
                self.loadings_learning_scores_ = self.loadings_learning_scores_[:-1]
                self.loadings_learning_rates_ = self.loadings_learning_rates_[:-1]

                # Remove memory
                del train_data, train_loader, _optimizer, _loadings, _intercept, _lr_scheduler
                if "cuda" in str(self.device):
                    torch.cuda.empty_cache()

                self._loadings_epochs = []
                self._intercept_epochs = []

                return self._saturated_loading_iter(
                    saturated_parameters=saturated_parameters,
                    X=X,
                )

        return (_loadings, _intercept)

    def _create_saturated_loading_optim(self, parameters: torch.Tensor, X: torch.Tensor):
        r"""Initialises the optimisation problem.

        Parameters
        ----------
        saturated_parameters : torch.Tensor
            Saturated parameters of the dataset X ($g^{-1}\left(X\right)$)
        X : torch.Tensor
            Dataset with cells in the rows and features in the columns.

        Returns
        -------
        optimizer: mctorch.optim
            mctorch optimiser instance

        loadings: mnn.Parameter
            mcTorch parameter instance with loadings constrained to the Stiefel manifold

        intercept: mnn.Parameter
            mcTorch parameter instance with intercept.

        lr_scheduler: torch.optim.scheduler
            Scheduler instance.
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Initialize loadings with spectrum (2**13 as maximum value for SVD to be relatively fast)
        random_batch_size = min(X.shape[0], 2**13)
        random_idx = np.random.choice(np.arange(parameters.shape[0]), replace=False, size=random_batch_size)
        if self.init == "spectral":
            _, _, v = torch.linalg.svd(parameters[random_idx] - torch.mean(parameters[random_idx], axis=0))
            loadings = mnn.Parameter(
                data=v[: self.n_pc, :].T, manifold=mnn.Stiefel(parameters.shape[1], self.n_pc), requires_grad=True
            )
        elif self.init == "random":
            loadings = mnn.Parameter(manifold=mnn.Stiefel(parameters.shape[1], self.n_pc), requires_grad=True)

        # Initialize intercept
        if self.exponential_family.family_name in ["poisson"]:
            intercept = mnn.Parameter(
                torch.median(parameters[random_idx], axis=0).values,
                manifold=mnn.Euclidean(parameters.shape[1]),
                requires_grad=True,
            )
        else:
            intercept = mnn.Parameter(
                torch.mean(parameters[random_idx], axis=0),
                manifold=mnn.Euclidean(parameters.shape[1]),
                requires_grad=True,
            )

        # Load ExponentialFamily params to GPU (if they exist)
        self.exponential_family.load_family_params_to_gpu(self.device)

        # Create optimizer
        # TODO: allow for other optimizer to be used.
        # TODO: learning rate for intercept.
        print("LEARNING RATE: %s" % (self.learning_rate_))
        optimizer = moptim.rAdagrad(
            params=[
                {"params": loadings, "lr": self.learning_rate_},
                {"params": intercept, "lr": self.learning_rate_ * 0.01},
            ]
        )
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=self.step_size, gamma=self.gamma)

        return optimizer, loadings, intercept, lr_scheduler

    def _optim_cost(
        self, loadings: torch.Tensor, intercept: torch.Tensor, batch_data: torch.Tensor, batch_parameters: torch.Tensor
    ):
        n = batch_data.shape[0]
        intercept_term = intercept.unsqueeze(0).repeat(n, 1)

        projected_parameters = batch_parameters - intercept_term
        projected_parameters = projected_parameters.matmul(loadings).matmul(loadings.T)
        projected_parameters = projected_parameters + intercept_term

        return self.exponential_family.log_likelihood(batch_data, projected_parameters)



import numpy as np
import umap
import pandas as pd
import scanpy as sc

from scanpy._utils import get_igraph_from_adjacency
import leidenalg as la

# own modules
from sincei.TopicModels import TOPICMODEL
from sincei.ParserCommon import numberOfProcessors

# umap.__version__ : should be >= 0.5.1

## literature on multi-graph clustering
"""
https://github.com/crisbodnar/regularised-spectral-clustering
https://arxiv.org/pdf/2103.16534.pdf
https://arxiv.org/pdf/2010.12301.pdf
https://arxiv.org/pdf/2010.15456.pdf
https://chenannie45.github.io/SDM18_MANE.pdf
https://www.researchgate.net/post/Python_or_R_packages_for_multilayer_network
http://www.mkivela.com/pymnet/visualizing.html#visualization-tutorial
https://cran.r-project.org/web/packages/multinet/multinet.pdf
https://sciendo.com/pdf/10.2478/amcs-2019-0010
https://web.media.mit.edu/~xdong/paper/tsp14.pdf
"""


def multiModal_clustering(
    mdata,
    modalities=None,
    method="PCA",
    modal_weights=None,
    column_key=None,
    nK=30,
    nPrinComps=20,
    clusterResolution=1.0,
    binarize=False,
    glmPCAfamily="poisson",
):
    r"""
    Performs multi-graph clustering on matched keys(barcodes) of a mudata object and stores the clustering results
    in mdata.obs["cluster_multi"]. It also stores the UMAP coordinates for each of the specified modalities in
    mdata[mod].obsm["X_umap"], where mod is the modality.
    Note: If method is "PCA" or "logPCA", the data matrix of the modality will be normalized, and log1p-transformed
    in the case of logPCA.

    Parameters
    ----------
    mudata : MuData
        MuData object containing several data modalities
    modalities : list[str]
        List of modalities to use for clustering, e.g. ["RNA", "ATAC", "CHiC"]
    method : list[str]
        What processing method to use for each modality. Choose between "PCA", "logPCA", "glmPCA", "LSA" or "LDA".
        Default is "PCA" for all modalities.
    modal_weights : list[float]
        Weights for each modality in the clustering process. Default is equal weighting. E.g. for RNA and CHiC, use [2, 1].
    column_key : str, optional
        Column name for the barcode. If None, the index of .obs for each modality is used.
    nK : int
        Number of nearest neighbours to consider for clustering and UMAP. This number should be chosen considering
        the total number of cells and expected number of clusters. Smaller number will lead to more fragmented clusters.
    nPrinComps : int or list[int]
        Number of principal components (for logPCA or glmPCA) or number of topics (for LSA and LDA) to use for model.
        Use higher number for samples with more expected heterogenity. If list is provided, it must contain a value for each
        modality. Default is 20.
    clusteResolution : float
        Resolution parameter for clustering. Values lower than 1.0 result in less clusters, while higher values lead to
        splitting of clusters. In most cases, the optimum value would be between 0.8 and 1.2. Default is 1.0 .
    binarize : bool
        Whether to binarize the counts per region before dimensionality reduction (only for LSA/LDA).
    glmPCAfamily : str
        The choice of exponential family distribution to use for glmPCA method. Default is "poisson".
    """

    # check if modalities are provided, otherwise use all
    if modalities is None:
        raise ValueError(f"Choose modalities to use for clustering.")
    # check if modalities are in mudata object
    for mod in modalities:
        if mod not in mdata.mod.keys():
            raise ValueError(f"Modality {mod} not found in MuData object.")
    # check if method is provided, otherwise use "PCA" for all
    if method == "PCA":
        method = ["PCA"] * len(modalities)
    # check if modalities and method lists have the same length
    if len(modalities) != len(method):
        raise ValueError(f"Modalities and method lists must have the same length.")
    # check if modal_weights are provided, otherwise use equal weights
    if modal_weights is None:
        modal_weights = [1] * len(modalities)
    # check if modal_weights and modalities lists have the same length
    if len(modal_weights) != len(modalities):
        raise ValueError(f"Modalities and modal_weights lists must have the same length.")

    # Find common barcodes in provided modalities
    if column_key is None:
        barcodes = set.intersection(*(set(mdata.mod[mod].obs.index) for mod in modalities))
    else:
        barcodes = set.intersection(*(set(mdata.mod[mod].obs[column_key]) for mod in modalities))
    barcodes = list(barcodes)

    adatas = []
    graphs = []
    for mod, met in zip(modalities, method):
        adata = mdata.mod[mod][barcodes]

        if met == "PCA":
            # if no method is provided, use PCA
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.pca(adata, nPrinComps)

        elif met == "logPCA":
            ## log1p+PCA using scanpy
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)
            sc.pp.pca(adata, nPrinComps)

        elif met == "LSA":
            ## LSA using gensim
            model_object = TOPICMODEL(
                adata,
                n_topics=nPrinComps,
                binarize=binarize,
                smart_code="lfu",
            )
            model_object.runLSA()
            adata.obsm["X_pca"] = model_object.get_cell_topic()

        elif met == "LDA":
            ## LDA using gensim
            model_object = TOPICMODEL(
                adata,
                n_topics=nPrinComps,
                binarize=binarize,
                n_passes=2,
                n_workers=numberOfProcessors("max"),
            )
            model_object.runLDA()
            adata.obsm["X_pca"] = model_object.get_cell_topic()

        elif met == "glmPCA":
            # import glmPCA (not imported on top due to special optional import of mctorch)
            from sincei.GLMPCA import GLMPCA

            ## glmPCA using mctorch
            model_object = GLMPCA(
                n_pc=nPrinComps,
                family=glmPCAfamily,
            )
            model_object.fit(adata)
            cell_pcs = model_object.saturated_loadings_.detach().numpy()

            ## update the anndata object
            adata.obsm["X_pca"] = np.asarray(cell_pcs)

        sc.pp.neighbors(adata, use_rep="X_pca", n_neighbors=nK)
        sc.tl.leiden(adata, resolution=clusterResolution)
        sc.tl.paga(adata)
        sc.pl.paga(adata, plot=False, threshold=0.1)
        sc.tl.umap(adata, min_dist=0.1, spread=5, init_pos="paga")

        # get graph
        graph = get_igraph_from_adjacency(adata.obsp["connectivities"], directed=True)

        adatas.append(adata)
        graphs.append(graph)

    # leiden multi-layer clustering
    optimiser = la.Optimiser()
    parts = []
    for graph in graphs:
        part = la.ModularityVertexPartition(graph)
        parts.append(part)

    optimiser.optimise_partition_multiplex(parts, layer_weights=modal_weights, n_iterations=-1)
    print("Detected clusters: ", set(parts[0].membership))
    mdata.obs["cluster_multi"] = parts[0].membership

    for mod, adata in zip(modalities, adatas):
        mdata.mod[mod] = adata


def umap_aligned(mdata, modalities=None, column_key=None, nK=30, distance_metric="euclidean"):
    r"""
    Aligns the UMAP embeddings of the selected modalities in a mudata object using the UMAP AlignedUMAP
    class and stores them in mdata[mod].obsm["X_umap_aligned"], where mod is the modality. This produces
    an aligned UMAP for each modality, since the alignment for each may be slightly different.

    Parameters
    ----------
    mudata : MuData
        MuData object containing several data modalities
    modalities : list[str]
        List of modalities to use for clustering, e.g. ["RNA", "ATAC", "CHiC"]
    column_key : str, optional
        Column name for the barcode. If None, the index of .obs for each modality is used.
    nK : int
        Number of nearest neighbors to use for UMAP
    distance_metric : str
        Distance metric to use for UMAP, e.g. "euclidean", "cosine", etc.
    """
    # check if modalities are provided, otherwise use all
    if modalities is None:
        raise ValueError(f"Choose modalities to use to align UMAP.")
    # check if modalities are in mudata object
    for mod in modalities:
        if mod not in mdata.mod.keys():
            raise ValueError(f"Modality {mod} not found in MuData object.")

    # Find common barcodes in provided modalities
    if column_key is None:
        barcodes = set.intersection(*(set(mdata.mod[mod].obs.index) for mod in modalities))
    else:
        barcodes = set.intersection(*(set(mdata.mod[mod].obs[column_key]) for mod in modalities))
    barcodes = list(barcodes)

    adatas = []
    umaps = []
    for mod in modalities:
        adata = mdata.mod[mod][barcodes]
        try:
            um = adata.obs[:, f"UMAP1_{mod}", f"UMAP2_{mod}"]
        except KeyError:
            raise KeyError(f"UMAP coordinates for modality {mod} not found. Please run UMAP first.")

        adatas.append(adata)
        umaps.append(um)

    # AlignedUMAP requires a mapping of relations between modalities.
    # In our case, the numerical index of each cell barcode to itself are the relations.
    relation_dict = {i: i for i in range(len(barcodes))}
    relation_dicts = [relation_dict.copy() for i in range(len(modalities) - 1)]

    # UMAP
    UA = umap.AlignedUMAP(
        spread=10,
        min_dist=0.01,
        n_neighbors=nK,
        metric=distance_metric,
        init="random",
        random_state=42,
    )
    aligned_umap = UA.fit(umaps, relations=relation_dicts)

    # Update the mudata object with the aligned UMAP coordinates
    for i, mod in enumerate(modalities):
        mdata[mod].obsm["X_umap_aligned"] = aligned_umap.embeddings_[i]



import torch
import numpy as np
import scipy
from tqdm import tqdm
from joblib import Parallel, delayed

saturation_eps = 10**-10


class ExponentialFamily:
    r"""Encodes an exponential family distribution using PyTorch autodiff structures.

    ExponentialFamily corresponds to the superclass providing a backbone for
    all exponential family.
    Each subclass should contain the following methods, defined based on the
    distribution of choice (same notation as [Mourragui et al, 2023]):
        - sufficient_statistics ($T$),
        - natural_parametrization ($\eta$),
        - log_partition ($A$).
        - invert_g ($g^{-1}$).
        - initialize_family_parameters: computes parameters used in other
        methods, e.g., gene-level dispersion for Negative Binomial.
    We added a "base_measure" for sake of completeness, but this method is not
    necessary for running GLM-PCA.
    The log-likelihood and exponential term are defined directly from the
    aforementionned methods.

    Parameters
    ----------
    family_name : str
        Name of the family.
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "base"
        self.family_params = family_params if family_params else {}

    def sufficient_statistics(self, X: torch.Tensor):
        return X

    def natural_parametrization(self, theta: torch.Tensor):
        return theta

    def log_partition(self, theta: torch.Tensor):
        return 0.0

    def base_measure(self, X: torch.Tensor):
        return torch.ones(size=X.shape)

    def invert_g(self, X: torch.Tensor):
        return X

    def exponential_term(self, X: torch.Tensor, theta: torch.Tensor):
        return torch.multiply(self.sufficient_statistics(X), self.natural_parametrization(theta))

    def distribution(self, X: torch.Tensor, theta: torch.Tensor):
        f = self.base_measure(X)
        expt = self.exponential_term(X, theta) - self.log_partition(theta)

        return torch.multiply(f, torch.exp(expt))

    def log_distribution(self, X: torch.Tensor, theta: torch.Tensor):
        f = self.base_measure(X)
        expt = self.exponential_term(X, theta) - self.log_partition(theta)

        return expt - torch.log(f)

    def log_likelihood(self, X: torch.Tensor, theta: torch.Tensor):
        """Computes negative log-likelihood between dataset X and parameters theta"""
        expt = self.exponential_term(X, theta) - self.log_partition(theta)
        return -torch.sum(expt)

    def load_family_params_to_gpu(self, device):
        self.family_params = {
            k: (
                self.family_params[k].to(device)
                if type(self.family_params[k]) is torch.Tensor
                else self.family_params[k]
            )
            for k in self.family_params
        }

    def initialize_family_parameters(self, X: torch.Tensor = None):
        """General method to initialize certain parameters (e.g. for Beta or Negative Binomial"""
        pass


class Gaussian(ExponentialFamily):
    r"""Gaussian with standard deviation one.

    GLMPCA with Gaussian as family is equivalent to the standard PCA.
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "gaussian"
        self.family_params = family_params if family_params else {}

    def sufficient_statistics(self, X: torch.Tensor):
        return X

    def natural_parametrization(self, theta: torch.Tensor):
        return theta

    def log_partition(self, theta: torch.Tensor):
        return torch.square(theta) / 2.0

    def base_measure(self, X: torch.Tensor):
        return torch.exp(-torch.square(X) / 2.0) / np.sqrt(2.0 * torch.pi)

    def invert_g(self, X: torch.Tensor):
        return X


class Bernoulli(ExponentialFamily):
    r"""Bernoulli distribution

    family_params of interest:
        - "max_val" (int) corresponding to the max value (replaces infinity).
        Empirically, values above 10 yield similar results.
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "bernoulli"
        self.family_params = family_params if family_params else {"max_val": 30}

    def sufficient_statistics(self, X: torch.Tensor):
        return X

    def natural_parametrization(self, theta: torch.Tensor):
        return theta

    def log_partition(self, theta: torch.Tensor):
        return torch.log(1.0 + torch.exp(theta))

    def base_measure(self, X: torch.Tensor):
        return 1.0

    def invert_g(self, X: torch.Tensor):
        return torch.log(X / (X - 1)).clip(-self.family_params["max_val"], self.family_params["max_val"])

    def log_likelihood(self, X: torch.Tensor, theta: torch.Tensor):
        """Computes negative log-likelihood between dataset X and parameters theta"""
        expt = self.exponential_term(X, theta) - self.log_partition(theta)
        return -torch.sum(expt)


class Poisson(ExponentialFamily):
    r"""Poisson distribution

    family_params of interest:
        - "min_val" (int) corresponding to the min value (replaces 0).
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "poisson"
        self.family_params = family_params if family_params else {"min_val": 1e-20}

    def sufficient_statistics(self, X: torch.Tensor):
        return X

    def natural_parametrization(self, theta: torch.Tensor):
        return theta

    def log_partition(self, theta: torch.Tensor):
        return torch.exp(theta)

    def base_measure(self, X: torch.Tensor):
        return 1 / scipy.special.gamma(X + 1)

    def invert_g(self, X: torch.Tensor):
        return torch.log(X).clip(self.family_params["min_val"])

    def log_distribution(self, X: torch.Tensor, theta: torch.Tensor):
        """The computation of gamma function for the base measure (h) would lead to inf, hence a re-design
        of the method."""
        log_f = torch.lgamma(X + 1)
        expt = self.exponential_term(X, theta) - self.log_partition(theta)

        return expt - log_f


class Beta(ExponentialFamily):
    r"""Beta distribution, using a standard formulation.

    Original formulation presented in [Mourragui et al, 2023].

    family_params of interest:
        - "min_val" (int): min data value (replaces 0 and 1).
        - "n_jobs" (int): number of jobs, specifically
        for computing the "nu" parameter.
        - "method" (str): method use to compute the "nu" parameter per
        feature. Two possibles: "MLE" and "MM". Default to "MLE".
        - "eps" (float): minimum difference used for inverting the g
        function. Default to 1e-4
        - "maxiter" (int): maximum number of iterations for the inversion
        of the g function. Default to 100.
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "beta"
        if family_params is None or "nu" not in family_params:
            print("Beta distribution not initialized yet")
        default_family_params = {"min_val": 1e-5, "n_jobs": 1, "eps": 1e-4, "maxiter": 100, "method": "MLE"}
        self.family_params = family_params if family_params else default_family_params
        for k in kwargs:
            self.family_params[k] = kwargs[k]
        for k in default_family_params:
            if k in self.family_params:
                continue
            self.family_params[k] = default_family_params[k]

    def sufficient_statistics(self, X: torch.Tensor):
        X = X.clip(self.family_params["min_val"], 1 - self.family_params["min_val"])
        return torch.stack([torch.log(X), torch.log(1 - X)])

    def natural_parametrization(self, theta: torch.Tensor):
        nat_params = torch.stack([theta * self.family_params["nu"], (1 - theta) * self.family_params["nu"]])
        if nat_params.shape[1] == 1:
            nat_params = nat_params.flatten()
        return nat_params

    def base_measure(self, X: torch.Tensor):
        return torch.mul(X, 1 - X)

    def log_partition(self, theta: torch.Tensor):
        numerator = torch.sum(torch.lgamma(self.natural_parametrization(theta)), axis=0)
        denominator = torch.lgamma(self.family_params["nu"])
        return numerator - denominator

    def exponential_term(self, X: torch.Tensor, theta: torch.Tensor):
        return torch.sum(torch.multiply(self.sufficient_statistics(X), self.natural_parametrization(theta)), axis=0)

    def _derivative_log_likelihood(self, X: torch.Tensor, theta: torch.Tensor):
        X = X.clip(self.family_params["eps"], 1 - self.family_params["eps"])
        return (
            torch.log(X / (1 - X))
            + torch.digamma((1 - theta) * self.family_params["nu"])
            - torch.digamma(theta * self.family_params["nu"])
        )

    def initialize_family_parameters(self, X: torch.Tensor = None):
        p = X.shape[1]

        def compute_beta_param(x):
            y = x[x > self.family_params["eps"]]
            y = y[y < 1 - self.family_params["eps"]]
            return scipy.stats.beta.fit(y, floc=0, fscale=1, method=self.family_params["method"])

        self.family_params["nu"] = torch.Tensor(
            Parallel(n_jobs=self.family_params["n_jobs"], batch_size=100, backend="threading")(
                delayed(compute_beta_param)(X[:, idx]) for idx in tqdm(range(p))
            )
        )
        self.family_params["nu"] = torch.sum(self.family_params["nu"][:, :2], axis=1)
        assert self.family_params["nu"].shape[0] == p

        return True

    def invert_g(self, X: torch.Tensor = None):
        """Dichotomy to find where derivative maxes out"""

        X = X.clip(self.family_params["eps"], 1 - self.family_params["eps"])

        # Initialize dichotomy parameters.
        min_val = torch.zeros(X.shape)
        max_val = torch.ones(X.shape)
        theta = (min_val + max_val) / 2

        llik = self._derivative_log_likelihood(X, theta)
        for idx in tqdm(range(self.family_params["maxiter"])):
            min_val[llik > 0] = theta[llik > 0]
            max_val[llik < 0] = theta[llik < 0]
            theta = (min_val + max_val) / 2
            llik = self._derivative_log_likelihood(X, theta)

            if torch.max(torch.abs(llik)) < self.family_params["eps"]:
                print("CONVERGENCE AFTER %s ITERATIONS" % (idx))
                break

        if idx <= self.family_params["maxiter"]:
            print("CONVERGENCE NOT REACHED")

        return theta


class SigmoidBeta(Beta):
    r"""Beta distribution re-parametrized using a Sigmoid.

    This distribution is similar to the previous Beta (which it
    inherits from) but the natural parameter is re-parametrized using
    a Sigmoid. This is shown expeerimentally to stabilize the
    optimisation by removing the ]0,1[ constraint.

    family_params of interest:
        - "min_val" (int): min data value (replaces 0 and 1).
        - "n_jobs" (int): number of jobs, specifically
        for computing the "nu" parameter.
        - "method" (str): method use to compute the "nu" parameter per
        feature. Two possibles: "MLE" and "MM". Default to "MLE".
        - "eps" (float): minimum difference used for inverting the g
        function. Default to 1e-4
        - "maxiter" (int): maximum number of iterations for the inversion
        of the g function. Default to 100.
    """

    def natural_parametrization(self, theta: torch.Tensor):
        nat_params = torch.stack(
            [torch.sigmoid(theta) * self.family_params["nu"], (1 - torch.sigmoid(theta)) * self.family_params["nu"]]
        )
        if nat_params.shape[1] == 1:
            nat_params = nat_params.flatten()
        return nat_params

    def _derivative_log_likelihood(self, X: torch.Tensor, logit_theta: torch.Tensor):
        X = X.clip(self.family_params["eps"], 1 - self.family_params["eps"])
        return (
            torch.log(X / (1 - X))
            + torch.digamma((1 - logit_theta) * self.family_params["nu"])
            - torch.digamma(logit_theta * self.family_params["nu"])
        )

    def invert_g(self, X: torch.Tensor = None):
        """Dichotomy to find where derivative maxes out"""

        X = X.clip(self.family_params["eps"], 1 - self.family_params["eps"])

        # Initialize dichotomy parameters.
        min_val = torch.zeros(X.shape)
        max_val = torch.ones(X.shape)
        logit_theta = (min_val + max_val) / 2

        llik = self._derivative_log_likelihood(X, logit_theta)
        for idx in tqdm(range(self.family_params["maxiter"])):
            min_val[llik > 0] = logit_theta[llik > 0]
            max_val[llik < 0] = logit_theta[llik < 0]
            logit_theta = (min_val + max_val) / 2
            llik = self._derivative_log_likelihood(X, logit_theta)

            if torch.max(torch.abs(llik)) < self.family_params["eps"]:
                print("CONVERGENCE AFTER %s ITERATIONS" % (idx))
                break

        if idx <= self.family_params["maxiter"]:
            print("CONVERGENCE NOT REACHED")

        return torch.logit(logit_theta)


class Gamma(ExponentialFamily):
    r"""Gamma distribution using a standard formulation.

    Original formulation presented in [Mourragui et al, 2023].

    family_params of interest:
        - "min_val" (int): min data value, default to 1e-5.
        - "max_val" (int): max data value, default to 10e6.
        - "n_jobs" (int): number of jobs, specifically
        for computing the "nu" parameter.
        - "method" (str): method use to compute the "nu" parameter per
        feature. Two possibles: "MLE" and "MM". Default to "MLE".
        - "eps" (float): minimum difference used for inverting the g
        function. Default to 1e-4
        - "maxiter" (int): maximum number of iterations for the inversion
        of the g function. Default to 100.
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "gamma"
        if family_params is None or "nu" not in family_params:
            print("Gamma distribution not initialized yet")
        default_family_params = {"min_val": 1e-5, "max_val": 10e6, "n_jobs": 1, "eps": 1e-4, "maxiter": 100}
        self.family_params = family_params if family_params else default_family_params
        for k in kwargs:
            self.family_params[k] = kwargs[k]

    def sufficient_statistics(self, X: torch.Tensor):
        return torch.stack([torch.log(X), X])

    def natural_parametrization(self, theta: torch.Tensor):
        nat_params = torch.stack([theta, -torch.ones(theta.shape) * self.family_params["nu"]])
        if nat_params.shape[1] == 1:
            nat_params = nat_params.flatten()
        return nat_params

    def log_partition(self, theta: torch.Tensor):
        return torch.lgamma(theta + 1) - (theta + 1) * torch.log(self.family_params["nu"])

    def exponential_term(self, X: torch.Tensor, theta: torch.Tensor):
        return torch.sum(torch.multiply(self.sufficient_statistics(X), self.natural_parametrization(theta)), axis=0)

    def _digamma_implicit_function(self, X: torch.Tensor, theta: torch.Tensor):
        return torch.digamma(theta + 1) - torch.log(X * self.family_params["nu"])

    def invert_g(self, X: torch.Tensor = None):
        """Dichotomy to compute inverse of digamma function."""

        # Initialize dichotomy parameters.
        min_val = torch.zeros(X.shape)
        max_val = torch.ones(X.shape) * self.family_params["max_val"]
        theta = (min_val + max_val) / 2

        llik = self._digamma_implicit_function(X, theta)
        for idx in tqdm(range(self.family_params["maxiter"])):
            max_val[llik > 0] = theta[llik > 0]
            min_val[llik < 0] = theta[llik < 0]
            theta = (min_val + max_val) / 2
            llik = self._digamma_implicit_function(X, theta)

            if torch.max(torch.abs(llik)) < self.family_params["eps"]:
                print("CONVERGENCE AFTER %s ITERATIONS" % (idx))
                break

        if idx == self.family_params["maxiter"]:
            print("CONVERGENCE NOT REACHED")

        return theta

    def initialize_family_parameters(self, X: torch.Tensor = None):
        p = X.shape[1]

        self.family_params["nu"] = torch.Tensor(
            Parallel(n_jobs=self.family_params["n_jobs"])(
                delayed(scipy.stats.gamma.fit)(X[:, idx], floc=0) for idx in tqdm(range(p))
            )
        )
        self.family_params["nu"] = 1.0 / self.family_params["nu"][:, -1]
        assert self.family_params["nu"].shape[0] == p

        return True


class LogNormal(ExponentialFamily):
    r"""Log-normal distribution using a standard formulation.

    Original formulation presented in [Mourragui et al, 2023].

    family_params of interest:
        - "min_val" (int): min data value, default to 1e-5.
        - "max_val" (int): max data value, default to 10e6.
        - "n_jobs" (int): number of jobs, specifically
        for computing the "nu" parameter.
        - "method" (str): method use to compute the "nu" parameter per
        feature. Two possibles: "MLE" and "MM". Default to "MLE".
        - "eps" (float): minimum difference used for inverting the g
        function. Default to 1e-4
        - "maxiter" (int): maximum number of iterations for the inversion
        of the g function. Default to 100.
    """

    def __init__(self, family_params=None, **kwargs):
        self.family_name = "log_normal"
        if family_params is None or "nu" not in family_params:
            print("Log Normal distribution not initialized yet")
        default_family_params = {
            "min_val": 1e-8,
            "max_val": 10e6,
            "n_jobs": 1,
            "eps": 1e-4,
            "maxiter": 100,
        }
        self.family_params = family_params if family_params else default_family_params
        for k in kwargs:
            self.family_params[k] = kwargs[k]

    def sufficient_statistics(self, X: torch.Tensor):
        log_X = torch.log(X)
        return torch.stack([log_X, torch.square(log_X)])

    def natural_parametrization(self, theta: torch.Tensor):
        nat_params = torch.stack(
            [
                theta / torch.square(self.family_params["nu"]),
                -torch.ones(theta.shape) / (2 * torch.square(self.family_params["nu"])),
            ]
        )
        if nat_params.shape[1] == 1:
            nat_params = nat_params.flatten()
        return nat_params

    def base_measure(self, X: torch.Tensor):
        return 1 / (np.sqrt(2 * torch.pi) * X)

    def log_partition(self, theta: torch.Tensor):
        return torch.square(theta) / (2 * torch.square(self.family_params["nu"])) + torch.log(self.family_params["nu"])

    def exponential_term(self, X: torch.Tensor, theta: torch.Tensor):
        return torch.sum(torch.multiply(self.sufficient_statistics(X), self.natural_parametrization(theta)), axis=0)

    def invert_g(self, X: torch.Tensor = None):
        return torch.log(X.clip(self.family_params["min_val"]))

    def initialize_family_parameters(self, X: torch.Tensor = None):
        self.family_params["nu"] = torch.sqrt(torch.var(torch.log(X), axis=0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

# single-cell stuff
import anndata as ad
import mudata as md
import scanpy as sc

from sincei import ParserCommon
from sincei.ParserCommon import smartLabel


def parseArguments():
    io_args = ParserCommon.inputOutputOptions(opts=["h5adfiles"])
    other_args = ParserCommon.otherOptions()

    parser = argparse.ArgumentParser(
        parents=[io_args, get_args(), other_args],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="""
        This tool combines multiple count matrices (output of scCountReads) into one, either assuming they are different samples (multi-sample)
        or different measurements on the same set of cells (multi-modal). The result is a .h5ad (AnnData) file with combined counts in
        multi-sample mode or a .h5mu (MuData) file in multi-modal mode. NOTE: it doesn't perform any 'batch effect correction' or 'integration'
        of data from different technologies, which requires more sophisticated methods.
        """,
        usage="Example usage: scCombineCounts -i sample1.h5ad sample2.h5ad -o combined.h5ad -m multi-sample > log.txt",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)

    general = parser.add_argument_group("General Options")

    general.add_argument(
        "--outFile",
        "-o",
        type=str,
        help="The file to write results to. For method: `multi-sample`, the output "
        "file is an updated .h5ad object, which can be used by other tools. "
        "For method: `multi-modal`, the output file is an .h5mu file. This file can only be "
        "used by scClusterCells, to perform multi-modal clustering. ",
        required=True,
    )

    general.add_argument(
        "--labels",
        "-l",
        metavar="sample1 sample2",
        help="User defined labels instead of default labels from "
        "file names. Multiple labels have to be separated by a space, e.g. "
        "--labels sample1 sample2 sample3. In case of --method 'multi-modal' "
        "the --labels can be used as the labels of different data modalities (eg. RNA, ATAC)",
        nargs="+",
    )

    general.add_argument(
        "--method",
        "-m",
        type=str,
        choices=["multi-sample", "multi-modal"],
        default="multi-sample",
        help="How to merge the counts from the provided samples. "
        "`multi-sample`: assumes that each sample is the independent, "
        "but were counted in the same manner (i.e. on same features), therefore "
        "it looks for feature overlaps, but not for barcode overlaps. "
        "`multi-modal`: assumes that the counts were generated in 2 different ways, "
        "but from the same set of cells (for example, using a multi-omic technology), "
        "therefore it looks for the overlap of cell barcodes, but not for the overlaps "
        "of features (Default: %(default)s)",
    )

    return parser


def main(args=None):
    args = parseArguments().parse_args(args)
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    if args.labels and len(args.input) != len(args.labels):
        print("The number of labels does not match the number of input files.")
        sys.exit(1)
    if not args.labels:
        # try smartlabel
        args.labels = [smartLabel(x) for x in args.input]
    adata_list = [sc.read_h5ad(x) for x in args.input]

    if args.method == "multi-sample":
        ## concatenate labels and match chrom, start, end
        var_list = []
        var_cols = ["chrom", "start", "end"]
        for lab, adata in zip(args.labels, adata_list):
            obs = adata.obs_names.to_list()
            lab = [lab] * len(obs)
            new_names = ["_".join([x, y]) for x, y in zip(lab, obs)]
            adata.obs_names = new_names
            hasinfo = all([x in adata.var.columns for x in var_cols])
            var_list.append(hasinfo)

        ## keep the chrom, start, end from original sample if present
        adata = ad.concat(adata_list)
        if all(var_list):
            var_df = adata_list[0].var[var_cols]
            adata.var = adata.var.join(var_df)
        else:
            sys.stderr.write(
                "WARNING: Not all input files contain the 'chrom', 'start', 'end' information. "
                "The output will lack these fields. This might cause an error in some downstream tools"
            )

        sys.stdout.write("Combined cells: {} \n".format(adata.shape[0]))
        sys.stdout.write("Combined features: {} \n".format(adata.shape[1]))
        adata.write_h5ad(args.outFile)

    elif args.method == "multi-modal":
        adata_dict = dict(zip(args.labels, adata_list))
        mdata = md.MuData(adata_dict)

        sys.stdout.write("Combined modalities: {} \n".format(len(mdata.mod)))
        sys.stdout.write("Combined cells: {} \n".format(mdata.shape[0]))
        sys.stdout.write("Combined features: {} \n".format(mdata.shape[1]))
        mdata.write_h5mu(args.outFile)

    else:
        sys.stderr.write("Choose a valid method: multi-sample or multi-modal.\n")
        sys.exit(1)

    return 0
